from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import json
import re
import numpy as np
import pandas as pd

from .data import _get_connection
from .plotting import _init_plot, _draw_plot
from .compound import Compound
from .element import Element

class Stack(object):
	"""Foil pack for stacked target calculations

	Computes the energy loss and (relative) charged particle flux through a stack
	of foils using the Anderson-Ziegler formulation for stopping powers.
	
	Parameters
	----------
	stack : list of dicts, pd.DataFrame or str
		Definition of the foils in the stack.  The 'compound' for each foil in
		the stack must be given, and the 'areal_density' or some combination of parameters
		that allow the areal density to be calculated must also be given.  Foils must
		also be given a 'name' if they are to be filtered by the .saveas(), .summarize(),
		and .plot() methods.  By default, foils without 'name' are not included by these
		methods.

		There are three acceptable formats for `stack`.  The first is a pd.DataFrame
		with the columns described. The second is a list of dicts, where each dict contains
		the appropriate keys.  The last is a str, which is a path to a file in either .csv,
		.json or .db format, where the headers of the file contain the correct information.
		Note that the .json file must follow the 'records' format (see pandas docs).  If a .db
		file, it must have a table named 'stack'.

		The 'areal_density' can be given directly, in units of mg/cm^2, or will be calculated
		from the following: 'mass' (in g) and 'area' (in cm^2), 'thickness' (in mm) and 'density'
		(in g/cm^3), or just 'thickness' if the compound is a natural element, or 
		is in `ci.COMPOUND_LIST` or the 'compounds' argument.

		Also, the following shorthand indices are supported: 'cm' for 'compound', 'd' for
		'density', 't' for 'thickness', 'm' for 'mass', 'a' for 'area', 'ad' for 'areal_density',
		and 'nm' for 'name'.


	particle : str
		Incident ion.  For light ions, options are 'p' (default), 'd', 't', 'a' for proton, 
		deuteron, triton and alpha, respectively.  Additionally, heavy ions can be
		specified either by element or isotope, e.g. 'Fe', '40CA', 'U', 'Bi-209'.For 
		light ions, the charge state is assumed to be fully stripped. For heavy ions
		the charge state is handled by a Bohr/Northcliffe parameterization consistent
		with the Anderson-Ziegler formalism.

	E0 : float
		Incident particle energy, in MeV.  If dE0 is not provided, it will
		default to 1 percent of E0.


	Other Parameters
	----------------
	compounds : str, pandas.DataFrame, list or dict
		Compound definitions for the compounds included in the foil stack.  If the compounds
		are not natural elements, or `ci.COMPOUND_LIST`, or if different weights or densities
		are required, they can be specified here. (Note specifying specific densities in the
		'stack' argument is probably more appropriate.)  Also, if the 'compound' name in the
		stack is a chemical formula, e.g. 'H2O', 'SrCO3', the weights can be inferred and 
		'compounds' doesn't need to be given.

		If compounds is a pandas DataFrame, it must have the columns 'compound', 'element', one of 
		'weight', 'atom_weight', or 'mass_weight', and optionally 'density'.  If a str, it must be
		a path to a .csv, .json or .db file, where .json files must be in the 'records' format and
		.db files must have a 'compounds' table.  All must have the above information.  For .csv 
		files, the compound only needs to be given for the first line of that compound definition.

		If compounds is a list, it must be a list of ci.Element or ci.Compound classes.  If it is a
		dict, it must have the compound names as keys, and weights as values, e.g. 
		{'Water':{'H':2, 'O':1}, 'Brass':{'Cu':-66,'Zn':-33}}

	dE0 : float
		1-sigma width of the energy distribution from which the initial
		particle energies are sampled, in MeV.  Default is to 1 percent of E0.

	N : int
		Number of particles to simulate. Default is 10000.

	dp : float
		Density multiplier.  dp is uniformly multiplied to all areal densities in the stack.  Default 1.0.

	chunk_size : int
		If N is large, split the stack calculation in to multiple "chunks" of size `chunk_size`. Default 1E7.

	accuracy : float
		Maximum allowed (absolute) error in the predictor-corrector method. Default 0.01.  If error is
		above `accuracy`, each foil in the stack will be solved with multiple steps, between `min_steps`
		and `max_steps`.

	min_steps : int
		The minimum number of steps per foil, in the predictor-corrector solver.  Default 2.

	max_steps : int
		The maximum number of steps per foil, in the predictor-corrector solver.  Default 50.

	Attributes
	----------
	stack : pandas.DataFrame
		'name', 'compound', 'areal_density', mean energy 'mu_E', and 1-sigma energy width 'sig_E'
		for each foil in the stack (energies in MeV).

	fluxes : pandas.DataFrame
		'flux' as a function of 'energy' for each foil in the stack where 'name' is not None.

	compounds : dict
		Dictionary with compound names as keys, and ci.Compound classes as values.

	Examples
	--------
	>>> stack = [{'cm':'H2O', 'ad':800.0, 'name':'water'},
				{'cm':'RbCl', 'density':3.0, 't':0.03, 'name':'salt'},
				{'cm':'Kapton', 't':0.025},
				{'cm':'Brass', 'm':3.5, 'a':1.0, 'name':'metal'}]
	>>> st = ci.Stack(stack, compounds='example_compounds.json')
	>>> st = ci.Stack(stack, compounds={'Brass':{'Cu':-66, 'Zn':-33}}, E0=60.0)
	>>> print(st.stack)
	    name compound  areal_density       mu_E     sig_E
	0  water      H2O         800.00  55.444815  2.935233
	1   salt     RbCl           9.00  50.668313  0.683532
	2    NaN   Kapton           3.55  50.612543  0.683325
	3  metal    Brass         350.00  49.159245  1.205481
	>>> st.saveas('stack_calc.csv')

	"""

	def __init__(self, stack, particle='p', E0=60.0, **kwargs):
		self._E0, self._particle = float(E0), particle
		self.compounds = {}
		self._parse_kwargs(**kwargs)

		if type(stack)==str:
			if stack.endswith('.json'):
				df = pd.read_json(stack, orient='records')

			elif stack.endswith('.csv'):
				df = pd.read_csv(stack, header=0)

			elif stack.endswith('.db'):
				df = pd.read_sql('SELECT * FROM stack', _get_connection(stack))

		elif type(stack)==list:
			df = pd.DataFrame(stack)

		elif type(stack)==pd.DataFrame:
			df = stack

		df = self._filter_cols(df)
		for cm in df['compound']:
			if cm not in self.compounds:
				self.compounds[cm] = Compound(cm)

		def _ad(s):
			if not np.isnan(s['areal_density']):
				return s['areal_density']
			if not np.isnan(s['mass']) and not np.isnan(s['area']):
				return 1E3*s['mass']/s['area']
			if not np.isnan(s['thickness']):
				if not np.isnan(s['density']):
					return 1E2*s['density']*s['thickness']
				else:
					return 1E2*self.compounds[s['compound']].density*s['thickness']

		ad = df.apply(_ad, axis=1)

		self.stack = pd.DataFrame({'name':df['name'], 'compound':df['compound'], 'areal_density':ad})[['name', 'compound', 'areal_density']]
		self._solve()


	def _filter_cols(self, df):
		cols = []
		for cl in df.columns:
			c = cl.lower()
			if c=='cm':
				cols.append('compound')
			elif c=='d':
				cols.append('density')
			elif c=='t':
				cols.append('thickness')
			elif c=='m':
				cols.append('mass')
			elif c=='a':
				cols.append('area')
			elif c=='ad':
				cols.append('areal_density')
			elif c=='nm':
				cols.append('name')
			else:
				cols.append(c)

		df.columns = cols
		for c in ['name','density','thickness','mass','area','areal_density']:
			if c not in df.columns:
				df[c] = np.nan
		return df


	def _parse_kwargs(self, **kwargs):
		self._dE0 = float(kwargs['dE0']) if 'dE0' in kwargs else 0.01*self._E0
		self._N = int(kwargs['N']) if 'N' in kwargs else 10000
		self._dp = float(kwargs['dp']) if 'dp' in kwargs else 1.0
		self._chunk_size = int(kwargs['chunk_size']) if 'chunk_size' in kwargs else int(1E7)
		self._accuracy = float(kwargs['accuracy']) if 'accuracy' in kwargs else 0.01
		self._min_steps = int(kwargs['min_steps']) if 'min_steps' in kwargs else 2
		self._max_steps = int(kwargs['max_steps']) if 'max_steps' in kwargs else 50

		if 'compounds' in kwargs:
			compounds = kwargs['compounds']
			if type(compounds)==str:
				if compounds.endswith('.json'):
					df = pd.read_json(compounds, orient='records').fillna(method='ffill')
					df.columns = map(str.lower, map(str, df.columns))
					cms = [str(i) for i in pd.unique(df['compound'])]
					self.compounds = {cm:Compound(cm, weights=df[df['compound']==cm]) for cm in cms}

				elif compounds.endswith('.csv'):
					df = pd.read_csv(compounds, header=0).fillna(method='ffill')
					df.columns = map(str.lower, map(str, df.columns))
					cms = [str(i) for i in pd.unique(df['compound'])]
					self.compounds = {cm:Compound(cm, weights=df[df['compound']==cm]) for cm in cms}

				elif compounds.endswith('.db'):
					df = pd.read_sql('SELECT * FROM compounds', _get_connection(compounds))
					df.columns = map(str.lower, map(str, df.columns))
					cms = [str(i) for i in pd.unique(df['compound'])]
					self.compounds = {cm:Compound(cm, weights=df[df['compound']==cm]) for cm in cms}


			elif type(compounds)==pd.DataFrame:
				compounds.columns = map(str.lower, map(str, compounds.columns))
				cms = [str(i) for i in pd.unique(compounds['compound'])]
				self.compounds = {cm:Compound(cm, weights=compounds[compounds['compound']==cm]) for cm in cms}

			elif type(compounds)==list:
				for cm in compounds:
					if type(cm)==Compound or type(cm)==Element:
						### ci.Compound class
						self.compounds[cm.name] = cm

			elif type(compounds)==dict:
				### e.g. {'Water':{'H':2, 'O':1}, 'Brass':{'Cu':-66,'Zn':-33}}
				self.compounds = {c:Compound(c, weights=compounds[c]) for c in compounds}



	def _solve_chunk(self, N):
		E0 = self._E0+self._dE0*np.random.normal(size=int(N))
		bins = np.arange(0.0, self._E0+10.0*self._dE0, min([0.1, self._E0/500.0]))

		hists = []
		dp = self._dp

		for n, sm in self.stack.iterrows():
			E_bar = [E0]

			if np.average(E0)<=0.0:
				hists.append(np.concatenate([[N],np.zeros(len(bins)-2)]))

			else:
				cm = self.compounds[sm['compound']]
				ad = sm['areal_density']
				steps = int((1.0/self._accuracy)*ad*dp*cm.S(np.average(E0), particle=self._particle, density=1E-3)/np.average(E0))
				steps = min([max([self._min_steps, steps]), self._max_steps])

				dr = (1.0/float(steps))
				for i in range(steps):
					S1 = cm.S(E0, particle=self._particle, density=1E-3)

					E1 = E0 - dr*dp*ad*S1
					E1 = np.where(E1>0, E1, 0.0)

					E0 = E0 - dr*0.5*dp*ad*(S1+cm.S(E1, particle=self._particle, density=1E-3))
					E0 = np.where(E0>0, E0, 0.0)

					E_bar.append(E0)

				hists.append(np.histogram(np.concatenate(E_bar), bins=bins)[0])

		return hists

	def _solve(self):
		dN = np.linspace(0, self._N, int(np.ceil(self._N/float(self._chunk_size)))+1, dtype=int)
		bins = np.arange(0.0, self._E0+10.0*self._dE0, min([0.1, self._E0/500.0]))
		energy = 0.5*(bins[1:]+bins[:-1])

		histos = list(map(self._solve_chunk, dN[1:]-dN[:-1]))

		self._flux_list = []
		warn = True
		mu_E, sig_E = [], []

		for n in range(len(self.stack)):
			sm = {}
			sm['flux'] = np.sum([h[n] for h in histos], axis=0)
			sm['flux'] = sm['flux']/np.sum(sm['flux'])

			mu_E.append(np.sum(sm['flux']*energy))
			sig_E.append(np.sqrt(np.sum(sm['flux']*(energy-mu_E[-1])**2)))

			lh = np.where(sm['flux']>0)[0]
			if lh.size:
				if lh[0]==0 and warn:
					print('WARNING: Beam stopped in foil {}'.format(n))
					warn = False

			if str(self.stack['name'][n])!='nan':
				sm['name'] = str(self.stack['name'][n])
				sm['flux'] = sm['flux'][lh[0]:lh[-1]+1]
				sm['energy'] = energy[lh[0]:lh[-1]+1]
				sm['bins'] = bins[lh[0]:lh[-1]+2]
				self._flux_list.append(sm)

		cols = ['name','energy','flux']
		self.fluxes = pd.concat([pd.DataFrame({c:sm[c] for c in cols}, columns=cols) for sm in self._flux_list], ignore_index=True)

		self.stack['mu_E'] = mu_E
		self.stack['sig_E'] = sig_E


	def _filter_samples(self, df, name):
		if name==False:
			return df

		df_NN = df[df['name'].notnull()]
		if name==True:
			return df_NN

		match = [bool(re.search(name, nm)) for nm in df_NN['name']]
		return df_NN[match]

	def get_flux(self, sample_name):
		"""Returns the computed energy grid and flux for a sample

		Units of energy are in MeV, and the flux is a relative flux (normalized to 1).
		Note that sample_name must be an exact match with the 'name' property in the
		stack specification.

		Parameters
		----------
		sample_name : str
			Name of the sample for which to return the flux.  Must be an exact
			match for an element in the stack.

		Returns
		-------
		(energy, flux) : tuple(np.ndarray, np.ndarray)
			Energy grid in MeV, and relative flux in the sample, along the energy grid.

		Examples
		--------
		>>> stack = [{'cm':'H2O', 'ad':800.0, 'name':'water'},
				{'cm':'RbCl', 'density':3.0, 't':0.03, 'name':'salt'},
				{'cm':'Kapton', 't':0.025},
				{'cm':'Brass','ad':350, 'name':'metal'}]
		>>> st = ci.Stack(stack, compounds={'Brass':{'Cu':-66, 'Zn':-33}}, E0=60.0)
		>>> print(st.get_flux('water'))
		(array([47.95, 48.05, 48.15, 48.25, 48.35, 48.45, 48.55, 48.65, 48.75,
		48.85, 48.95, 49.05, 49.15...
		>>> rx = ci.Reaction('16O(p,x)16F')
		>>> print(rx.average(*st.get_flux('water')))
		1.495163043176288

		"""

		df = self._filter_samples(self.fluxes, sample_name)
		match = [sample_name==nm for nm in df['name']]
		df = df[match]
		return df['energy'].to_numpy(), df['flux'].to_numpy()

	def saveas(self, filename, save_fluxes=True, filter_name=True):
		"""Saves the results of the stack calculation

		Saves the stack design, mean energies, and (optionally) the flux
		profile for each foil in the stack.  Supported file types are '.csv',
		'.db' and '.json'.

		Parameters
		----------
		filename : str
			Name of file to save to.  Supported file types are '.csv',
			'.db' and '.json'. If `save_fluxes=True`, an additional file
			will be saved to 'fname_fluxes.ftype'.

		save_fluxes : bool, optional
			If True, an additional file will be saved with the flux profile
			for each foil in the stack.  The foil must have a 'name' keyword,
			and can be further filtered with the `filter_name` argument.  If 
			false, only the stack design and mean energies are saved. Defaut, True.

		filter_name : bool or str, optional
			Applies a filter to the stack and fluxes before saving.  If True, only
			foils with a 'name' keyword will be saved. If 'False', foils without
			a 'name' will be saved in the stack design file, but not the fluxes
			file.  If a str, foils with a 'name' matching a regex search with filter_name
			are saved.  This applies to both files. Default, True.

		Examples
		--------
		>>> stack = [{'cm':'H2O', 'ad':800.0, 'name':'water'},
				{'cm':'RbCl', 'density':3.0, 't':0.03, 'name':'salt'},
				{'cm':'Kapton', 't':0.025},
				{'cm':'Brass','ad':350, 'name':'metal'}]

		>>> st = ci.Stack(stack, compounds={'Brass':{'Cu':-66, 'Zn':-33}}, E0=60.0)
		>>> st.saveas('example_stack.csv')
		>>> st.saveas('example_stack.json', filter_name=False)
		>>> st.saveas('example_stack.db', save_fluxes=False)

		"""

		if any([filename.endswith(e) for e in ['.png','.pdf','.eps','.pgf','.ps','.raw','.rgba','.svg','.svgz']]):
			self.plot(saveas=filename, show=False)

		if filename.endswith('.csv'):
			self._filter_samples(self.stack, filter_name).to_csv(filename, index=False)
			if save_fluxes:
				self._filter_samples(self.fluxes, filter_name).to_csv(filename.replace('.csv','_fluxes.csv'), index=False)

		if filename.endswith('.db'):
			self._filter_samples(self.stack, filter_name).to_sql('stack', _get_connection(filename), if_exists='replace', index=False)
			if save_fluxes:
				self._filter_samples(self.fluxes, filter_name).to_sql('fluxes', _get_connection(filename), if_exists='replace', index=False)

		if filename.endswith('.json'):
			json.dump(json.loads(self._filter_samples(self.stack, filter_name).to_json(orient='records')), open(filename, 'w'), indent=4)
			if save_fluxes:
				json.dump(json.loads(self._filter_samples(self.fluxes, filter_name).to_json(orient='records')), open(filename.replace('.json','_fluxes.json'), 'w'), indent=4)

		
	def summarize(self, filter_name=True):
		"""Summarize the stack calculation

		Prints out the mean energies and 1-sigma energy widths of
		each foil in the stack, or the filtered stack depending
		on the behavior of `filter_name`.

		Parameters
		----------
		filter_name : bool or str, optional
			Applies a filter to the stack.  If True, only
			foils with a 'name' keyword will be included. If 'False', a summary
			of all foils will be printed.  If a str, foils with a 'name' 
			matching a regex search with filter_name are included. Default, True.

		Examples
		--------
		>>> stack = [{'cm':'H2O', 'ad':800.0, 'name':'water'},
				{'cm':'RbCl', 'density':3.0, 't':0.03, 'name':'salt'},
				{'cm':'Kapton', 't':0.025},
				{'cm':'Brass','ad':350, 'name':'metal'}]

		>>> st = ci.Stack(stack, compounds={'Brass':{'Cu':-66, 'Zn':-33}}, E0=60.0)
		>>> st.summarize()
		water: 55.45 +/- 2.94 (MeV)
		salt: 50.68 +/- 0.69 (MeV)
		metal: 49.17 +/- 1.21 (MeV)
		>>> st.summarize(filter_name=False)
		water: 55.45 +/- 2.94 (MeV)
		salt: 50.68 +/- 0.69 (MeV)
		Kapton-1: 50.62 +/- 0.69 (MeV)
		metal: 49.17 +/- 1.21 (MeV)

		"""

		m = {}
		for n,sm in self._filter_samples(self.stack, filter_name).iterrows():
			if sm['name'] is None or str(sm['name'])=='nan':
				if sm['compound'] not in m:
					m[sm['compound']] = 1
				nm = sm['compound']+'-'+str(m[sm['compound']])
				m[sm['compound']] += 1

			else:
				nm = sm['name']
			print(nm+': '+str(round(sm['mu_E'], 2))+' +/- '+str(round(sm['sig_E'], 2))+' (MeV)')

		
	def plot(self, filter_name=None, **kwargs):
		"""Plots the fluxes for each foil in the stack calculation

		Plots the flux distribution for each foil in the stack, or
		the filtered stack depending on the behaviour of `filter_name`.

		Parameters
		----------
		filter_name : str, optional
			Applies a filter to the fluxes before plotting. If a str, 
			foils with a 'name' matching a regex search with filter_name
			are plotted.  Default, None.

		Other Parameters
		----------------
		**kwargs
			Optional keyword arguments for plotting.  See the 
			plotting section of the curie API for a complete
			list of kwargs.

		Examples
		--------
		>>> stack = [{'cm':'H2O', 'ad':800.0, 'name':'water'},
				{'cm':'RbCl', 'density':3.0, 't':0.03, 'name':'salt'},
				{'cm':'Kapton', 't':0.025},
				{'cm':'Brass','ad':350, 'name':'metal'}]

		>>> st = ci.Stack(stack, compounds={'Brass':{'Cu':-66, 'Zn':-33}}, E0=60.0)
		>>> st.plot()
		>>> st.plot(filter_name='salt')

		"""

		f,ax = _init_plot(**kwargs)
		filter_name = True if filter_name is None else filter_name
		for sm in self._flux_list:
			if type(filter_name)==str:
				if not re.search(filter_name, sm['name']):
					continue

			eng = np.array([sm['bins'][:-1], sm['bins'][1:]]).T.flatten()
			phi = np.array([sm['flux'], sm['flux']]).T.flatten()
			ax.plot(eng, phi, label=sm['name'])

		ax.set_xlabel('Energy (MeV)')
		ax.set_ylabel('Flux (a.u.)')
		ax.legend(loc=0)
		return _draw_plot(f, ax, **kwargs)
		


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re
import numpy as np
import pandas as pd

from scipy.interpolate import interp1d

from .data import _get_connection
from .plotting import _init_plot, _draw_plot
from .isotope import Isotope

ELEMENTS = ['n','H','He','Li','Be','B','C','N','O','F','Ne','Na','Mg','Al',
			'Si','P','S','Cl','Ar','K','Ca','Sc','Ti','V','Cr','Mn','Fe','Co',
			'Ni','Cu','Zn','Ga','Ge','As','Se','Br','Kr','Rb','Sr','Y','Zr',
			'Nb','Mo','Tc','Ru','Rh','Pd','Ag','Cd','In','Sn','Sb','Te','I',
			'Xe','Cs','Ba','La','Ce','Pr','Nd','Pm','Sm','Eu','Gd','Tb','Dy',
			'Ho','Er','Tm','Yb','Lu','Hf','Ta','W','Re','Os','Ir','Pt','Au',
			'Hg','Tl','Pb','Bi','Po','At','Rn','Fr','Ra','Ac','Th','Pa','U',
			'Np','Pu','Am','Cm','Bk','Cf','Es','Fm','Md','No','Lr','Rf','Db',
			'Sg','Bh','Hs','Mt','Ds','Rg','Cn','Nh','Fl','Mc','Lv','Ts']


class Element(object):
	"""Elemental data and properties

	The Element class provides useful data about the natural elements,
	such as mass, density, and isotopic composition.  Additionally, it contains
	functions for determining the interaction of radiation with the natural
	elements.  Principally it provides the mass-attenuation of photons, and 
	the stopping power/ranges of charged particles.
	
	Parameters
	----------
	element : str
		Symbol for the element, e.g. 'H', 'In', 'Zn', 'Fe'.  Case insensitive.
		Note that 'n' ("neutron") is not considered a valid element in this
		context, and will be interpreted as 'N' ("nitrogen").

	Attributes
	----------
	name : str
		Symbol for the element, in title-case. E.g. if input was 'fe', name 
		will be 'Fe'.

	Z : int
		Atomic number of the element.

	mass : float
		Molar mass of the natural element in atomic mass units (amu).

	isotopes : list
		List of isotopes with non-zero natural abundance.

	abundances : pd.DataFrame
		Natural abundances, in percent, for all isotopes found in nature. 
		Structure is a DataFrame with the columns 'isotope', 'abundance', 
		and 'unc_abundance'.

	density : float
		Density of the natural element in g/cm^3. The density is used in
		calculations of charged particle dEdx and photon attenuation, so
		you can assign a new density using `el.density = new_density` if
		needed, or using the `density` keyword in either of those functions.

	mass_coeff : pd.DataFrame
		Table of mass-attenuation coefficients as a function of photon
		energy, from the NIST XCOM database.  Energies are in keV, and
		mass-attenuation coefficients, or mu/rho, are given in cm^2/g.
		DataFrame columns are 'energy', 'mu' and 'mu_en' for the 
		mass-energy absorption coefficient.

	Examples
	--------
	>>> el = ci.Element('Fe')
	>>> print(el.mass)
	55.847
	>>> print(el.density)
	7.866

	"""

	def __init__(self, element):
		self.name = element.title()
		self.Z = ELEMENTS.index(self.name)

		df = pd.read_sql('SELECT * FROM weights WHERE Z={}'.format(self.Z), _get_connection('ziegler'))
		self.mass = df['amu'][0]
		self.density = df['density'][0]

		self.abundances = pd.read_sql('SELECT isotope, abundance, unc_abundance FROM chart WHERE Z={} AND abundance>0'.format(self.Z), _get_connection('decay'))
		if self.name=='Ta': ### Ta has a naturally occuring isomer
			self.abundances['isotope'] = map(lambda i:i.replace('180TA','180TAm1'), self.abundances['isotope'])
		self.isotopes = list(map(str, self.abundances['isotope']))

		self.mass_coeff = pd.read_sql('SELECT energy, mu, mu_en FROM mass_coeff WHERE Z={}'.format(self.Z), _get_connection('ziegler'))
		self._mc_interp, self._mc_en_interp = None, None
		self._A_p, self._A_he, self._I_p = None, None, None


	def __str__(self):
		return self.name

	def mu(self, energy):
		"""Mass-attenuation coefficient

		Interpolates the mass-attenuation coefficient, mu/rho,
		for the element along the input energy grid.

		Parameters
		----------
		energy : array_like
			The incident photon energy, in keV.

		Returns
		-------
		mu : np.ndarray
			Mass attenuation coefficient, mu/rho, in cm^2/g.

		Examples
		--------
		>>> el = ci.Element('Hg')
		>>> print(el.mu(200))
		0.9456

		"""

		if self._mc_interp is None:
			self._mc_interp = interp1d(np.log(self.mass_coeff['energy']), np.log(self.mass_coeff['mu']), bounds_error=False, fill_value='extrapolate')
		return np.exp(self._mc_interp(np.log(energy)))

	def mu_en(self, energy):
		"""Mass energy-absorption coefficient

		Interpolates the mass-energy absorption coefficient, mu_en/rho,
		for the element along the input energy grid.

		Parameters
		----------
		energy : array_like
			The incident photon energy, in keV.

		Returns
		-------
		mu_en : np.ndarray
			Mass energy absorption coefficient, mu_en/rho, in cm^2/g.

		Examples
		--------
		>>> el = ci.Element('Hg')
		>>> print(el.mu_en(200))
		0.5661

		"""

		if self._mc_en_interp is None:
			self._mc_en_interp = interp1d(np.log(self.mass_coeff['energy']), np.log(self.mass_coeff['mu_en']), bounds_error=False, fill_value='extrapolate')
		return np.exp(self._mc_en_interp(np.log(energy)))

	def attenuation(self, energy, x, density=None):
		"""Photon attenuation in matter

		Calculate the attenuation factor I(x)/I_0 = e^(-mu*x) for a given
		photon energy (in keV) and slab thickness (in cm).

		Parameters
		----------
		energy : array_like
			Incident photon energy in keV.

		x : float
			Thickness of slab of given element, in cm. 

		density : float, optional
			Density of the element in g/cm^3.  Default behavior is to
			use `Element.density`.

		Returns
		-------
		attenuation : numpy.ndarray
			The slab attenuation factor as an absolute number (i.e. from 0 to 1).
			E.g. if the incident intensity is I_0, the transmitted intensity I(x) 
			is I_0 times the attenuation factor.

		Examples
		--------
		>>> el = ci.Element('Fe')
		>>> print(el.attenuation(511, x=0.3))
		0.821621630674751
		>>> print(el.attenuation(300, x=0.5, density=8))
		0.6442940871813587

		"""

		energy = np.asarray(energy, dtype=np.float64)
		x = np.asarray(x, dtype=np.float64)
		if density is None:
			density = self.density

		return np.exp(-self.mu(energy)*x*density)


	def _parse_particle(self, particle):
		### Return Z, mass of incident particle
		if particle in ['p', 'd', 't', 'a']:
			return {'p':(1, 1.008), 'd':(1, 2.014), 't':(1, 3.016), 'a':(2, 4.002)}[particle]

		if particle=='P':
			print("WARNING: Assumed particle type P (phosphorus). If proton, use particle='p' (lower case).")
			return 15, 30.974

		### check if element or isotope
		if ''.join(re.findall('[A-Z]+', particle.upper())).title()==particle.title():
			el = Element(particle.title())
			return el.Z, el.mass

		else:
			ip = Isotope(particle)
			return ip.Z, ip.mass

	def _S_nucl(self, eng, z1, m1):
		RM = (m1+self.mass)*np.sqrt((z1**(2/3.0)+self.Z**(2/3.0)))
		ER = 32.53*self.mass*1E3*eng/(z1*self.Z*RM)

		return (0.5*np.log(1.0+ER)/(ER+0.10718+ER**0.37544))*8.462*z1*self.Z*m1/RM

	def _S_p(self, eng, M1=1.00727647):
		S = np.zeros(len(eng), dtype=np.float64) if eng.shape else np.array(0.0)
		E = 1E3*eng/M1

		if self._A_p is None:
			self._A_p = pd.read_sql('SELECT * FROM protons WHERE Z={}'.format(self.Z), _get_connection('ziegler')).to_numpy()[0,1:]
		A = self._A_p

		beta_sq = np.where(E>=1E3, 1.0-1.0/(1.0+E/931478.0)**2, 0.9)
		B0 = np.where(E>=1E3, np.log(A[6]*beta_sq/(1.0-beta_sq))-beta_sq, 0.0)
		Y = np.log(E[(E>=1E3)&(E<=5E4)])
		B0[np.where((E>=1E3)&(E<=5E4), B0, 0)!=0] -= A[7]+A[8]*Y+A[9]*Y**2+A[10]*Y**3+A[11]*Y**4

		S[E>=1E3] = (A[5]/beta_sq[E>=1E3])*B0[E>=1E3]

		S_low = A[1]*E[(E>=10)&(E<1E3)]**0.45
		S_high = (A[2]/E[(E>=10)&(E<1E3)])*np.log(1.0+(A[3]/E[(E>=10)&(E<1E3)])+A[4]*E[(E>=10)&(E<1E3)])

		S[(E>=10)&(E<1E3)] = S_low*S_high/(S_low+S_high)
		S[(E>0)&(E<10)] = A[0]*E[(E>0)&(E<10)]**0.5
		return S

	def _S_He(self, eng, M1=4.003):
		S = np.zeros(len(eng), dtype=np.float64) if eng.shape else np.array(0.0)
		E = eng*4.0015/M1
		E = np.where(E>=0.001, E, 0.001)

		if self._A_he is None:
			self._A_he = pd.read_sql('SELECT * FROM helium WHERE Z={}'.format(self.Z), _get_connection('ziegler')).to_numpy()[0,1:]
		A = self._A_he

		S_low = A[0]*(1E3*E[E<=10])**A[1]
		S_high = (A[2]/E[E<=10])*np.log(1.0+(A[3]/E[E<=10])+A[4]*E[E<=10])
		S[E<=10] = S_low*S_high/(S_low+S_high)

		Y = np.log(1.0/E[E>10])
		S[E>10] = np.exp(A[5]+A[6]*Y+A[7]*Y**2+A[8]*Y**3)
		return S

	def _S_elec(self, eng, z1, M1):
		S = np.zeros(len(eng), dtype=np.float64) if eng.shape else np.array(0.0)
		E_keV = 1E3*eng

		S[E_keV/M1<1000] = self._eff_Z_ratio(E_keV[E_keV/M1<1000], z1, M1)**2*self._S_p(eng[E_keV/M1<1000], M1)

		Y = E_keV[E_keV/M1>=1000]/M1
		beta_sq = 1.0-1.0/(1.0+Y/931478.0)**2
		FX = np.log(2E6*0.511003*beta_sq/(1.0-beta_sq))-beta_sq

		ZHY = 1.0-np.exp(-0.2*np.sqrt(Y)-0.0012*Y-0.00001443*Y**2)
		Z1EFF = self._eff_Z_ratio(E_keV[E_keV/M1>=1000], z1, M1)*ZHY

		if self._I_p is None:
			self._I_p = pd.read_sql('SELECT Ip_solid FROM ionization WHERE Z={}'.format(self.Z), _get_connection('ziegler'))['Ip_solid'][0]

		S[E_keV/M1>=1000] = 4E-1*np.pi*(1.9732857/137.03604)**2*Z1EFF**2*self.Z*(FX-np.log(self._I_p))/(0.511003*beta_sq)
		return S

	def _eff_Z_ratio(self, E_keV, z1, M1):
		if z1==1:
			return np.ones(len(eng))

		elif z1==2:
			Y = np.log(E_keV/M1)
			return z1*(1.0-np.exp(-0.7446-0.1429*Y-0.01562*Y**2+0.00267*Y**3-0.000001325*Y**8))

		elif z1==3:
			Y = E_keV/M1
			return z1*(1.0-np.exp(-0.7138-0.002797*Y-0.000001348*Y**2))

		BB = -0.886*np.sqrt(0.04*E_keV/M1)/z1**(2/3.0)
		return z1*(1.0-np.exp(BB-0.0378*np.sin(0.5*np.pi*BB))*(1.034-0.1777*np.exp(-0.08114*z1)))
		
	def S(self, energy, particle='p', density=None):
		"""Charged particle stopping power in matter

		Calculate the stopping power, S=-dE/dx, for a given ion as a 
		function of the ion energy in MeV.  Units of S are MeV/cm.  To return
		stopping power in units of MeV/(mg/cm^2), use option `density=1E-3`.

		Parameters
		----------
		energy : array_like
			Incident ion energy in MeV.

		particle : str, optional
			Incident ion.  For light ions, options are 'p' (default), 'd', 't', 'a' for proton, 
			deuteron, triton and alpha, respectively.  Additionally, heavy ions can be
			specified either by element or isotope, e.g. 'Fe', '40CA', 'U', 'Bi-209'. For
			light ions, the charge state is assumed to be fully stripped. For heavy ions
			the charge state is handled by a Bohr/Northcliffe parameterization consistent
			with the Anderson-Ziegler formalism.

		density : float, optional
			Density of the element in g/cm^3.  Default behavior is to use
			`Element.density`.  To return stopping power in units of MeV/(mg/cm^2), i.e.
			the mass-stopping power, use `density=1E-3`.

		Returns
		-------
		stopping_power : numpy.ndarray
			Stopping power, S=-dE/dx, for a given ion as a function of the 
			ion energy in MeV.  Units of S are MeV/cm.

		Examples
		--------
		>>> el = ci.Element('La')
		>>> print(el.S(60.0))
		36.8687750516453
		>>> print(el.S(55.0, density=1E-3)) ### S in MeV/(mg/cm^2)
		0.006371657662505643

		"""

		energy = np.asarray(energy, dtype=np.float64)
		if density is None:
			density = self.density

		Z_p, mass_p = self._parse_particle(particle)

		S = self._S_nucl(energy, Z_p, mass_p)
		if Z_p==1:
			S += self._S_p(energy, mass_p)
		elif Z_p==2:
			S += self._S_He(energy, mass_p)
		else:
			S += self._S_elec(energy, Z_p, mass_p)

		return (S*0.6022140857/self.mass)*1E3*density
		
	def range(self, energy, particle='p', density=None):
		"""Charged particle range in matter

		Calculates the charged particle range in the element, in cm.  Incident
		energy should be in MeV, and the particle type definition is identical
		to `Element.S()`.

		Parameters
		----------
		energy : array_like
			Incident ion energy in MeV.

		particle : str, optional
			Incident ion.  For light ions, options are 'p' (default), 'd', 't', 'a' for proton, 
			deuteron, triton and alpha, respectively.  Additionally, heavy ions can be
			specified either by element or isotope, e.g. 'Fe', '40CA', 'U', 'Bi-209'. For
			light ions, the charge state is assumed to be fully stripped. For heavy ions
			the charge state is handled by a Bohr/Northcliffe parameterization consistent
			with the Anderson-Ziegler formalism.

		density : float, optional
			Density of the element in g/cm^3.  Default behavior is to use
			`Element.density`.
		
		Returns
		-------
		range : np.ndarray
			Charged particle range in the element, in cm.

		Examples
		--------
		>>> el = ci.Element('Fe')
		>>> print(el.range(60.0))
		0.5858151125192633
		>>> el = ci.Element('U')
		>>> print(el.range(60.0))
		0.3763111404628591

		"""

		energy = np.asarray(energy, dtype=np.float64)
		
		dE = np.max(energy)/1E3
		E_min = min((np.min(energy), 1E-3))
		E_grid = np.arange(E_min, np.max(energy)+dE, dE)

		S = self.S(E_grid, particle=particle, density=density)
		x = np.cumsum((1.0/S)*dE)
		return interp1d(np.log(E_grid), x, bounds_error=None, fill_value='extrapolate')(np.log(energy))

		
	def plot_mass_coeff(self, energy=None, **kwargs):
		"""Plot the mass-attenuation coefficient in the element

		Creates a plot of the mass-attenuation coefficient (in cm^2/g)
		as a function of photon energy in keV.

		Parameters
		----------
		energy : array_like, optional
			Energy grid on which to plot, replacing the default energy grid.
			Units are in keV.
		
		Other Parameters
		----------------
		**kwargs
			Optional keyword arguments for plotting.  See the 
			plotting section of the curie API for a complete
			list of kwargs.

		Examples
		--------
		>>> el = ci.Element('Fe')
		>>> el.plot_mass_coeff()
		>>> el.plot_mass_coeff(style='poster')

		"""

		if energy is None:
			energy, mu = self.mass_coeff['energy'], self.mass_coeff['mu']
			
		else:
			energy = np.asarray(energy, dtype=np.float64)
			mu = self.mu(energy)

		f,ax = _init_plot(**kwargs)

		ax.plot(energy, mu, label=r'$\mu/\rho$'+' ({0}, Z={1})'.format(self.name, self.Z))
		ax.set_xlabel('Photon Energy (keV)')
		ax.set_ylabel(r'Attenuation Coeff. (cm$^2$/g)')
		ax.set_xscale('log')
		ax.set_yscale('log')
		ax.legend()
		
		return _draw_plot(f, ax, **kwargs)
		
	def plot_mass_coeff_en(self, energy=None, **kwargs):
		"""Plot the mass energy-absorption coefficient in the element

		Creates a plot of the mass energy-absorption coefficient (in cm^2/g)
		as a function of photon energy in keV.

		Parameters
		----------
		energy : array_like, optional
			Energy grid on which to plot, replacing the default energy grid.
			Units are in keV.

		Other Parameters
		----------------
		**kwargs
			Optional keyword arguments for plotting.  See the 
			plotting section of the curie API for a complete
			list of kwargs.

		Examples
		--------
		>>> el = ci.Element('Hf')
		
		Example plotting the mass-attenuation coefficient together with the mass
		energy-absorption coefficient, on the same axes.

		>>> f,ax = el.plot_mass_coeff(return_plot=True)
		>>> el.plot_mass_coeff_en(f=f, ax=ax)

		"""

		if energy is None:
			energy, mu = self.mass_coeff['energy'], self.mass_coeff['mu_en']
			
		else:
			energy = np.asarray(energy, dtype=np.float64)
			mu = self.mu_en(energy)

		f,ax = _init_plot(**kwargs)

		ax.plot(energy, mu, label=r'$\mu_{en}/\rho$'+' ({0}, Z={1})'.format(self.name, self.Z))
		ax.set_xlabel('Photon Energy (keV)')
		ax.set_ylabel(r'Attenuation Coeff. (cm$^2$/g)')
		ax.set_xscale('log')
		ax.set_yscale('log')
		ax.legend()

		return _draw_plot(f, ax, **kwargs)
		
	def plot_S(self, particle='p', energy=None, **kwargs):
		"""Plot the stopping power in the element

		Creates a plot of the charged particle stopping power (in MeV/(mg/cm^2))
		in the element as a function of the incident ion energy (in MeV).

		Parameters
		----------
		particle : str
			Incident ion.  For light ions, options are 'p' (default), 'd', 't', 'a' for proton, 
			deuteron, triton and alpha, respectively.  Additionally, heavy ions can be
			specified either by element or isotope, e.g. 'Fe', '40CA', 'U', 'Bi-209'.

		energy : array_like, optional
			Energy grid on which to plot, replacing the default energy grid.
			Units are in MeV.

		Other Parameters
		----------------
		**kwargs
			Optional keyword arguments for plotting.  See the 
			plotting section of the curie API for a complete
			list of kwargs.

		Examples
		--------
		>>> el = ci.Element('He')
		>>> el.plot_S(particle='a')
		>>> el = ci.Element('Fe')
		>>> el.plot_S(particle='d')


		"""

		if energy is None:
			energy = 10.0**np.arange(-1.5, 2.8, 0.05)

		f,ax = _init_plot(**kwargs)
		ax.plot(energy, self.S(energy, particle=particle, density=1E-3), label='S ({0}, Z={1})'.format(self.name, self.Z))

		ax.set_xlabel('Incident Energy (MeV)')
		ax.set_ylabel(r'Stopping Power (MeV/(mg/cm$^2$))')
		ax.set_xscale('log')
		ax.legend()

		return _draw_plot(f, ax, **kwargs)
		
	def plot_range(self, particle='p', energy=None, density=None, **kwargs):
		"""Plot the charged particle range in the element

		Creates a plot of the charged particle range (in cm)
		in the element as a function of the incident ion energy (in MeV).

		Parameters
		----------
		particle : str
			Incident ion.  For light ions, options are 'p' (default), 'd', 't', 'a' for proton, 
			deuteron, triton and alpha, respectively.  Additionally, heavy ions can be
			specified either by element or isotope, e.g. 'Fe', '40CA', 'U', 'Bi-209'.

		energy : array_like, optional
			Energy grid on which to plot, replacing the default energy grid.
			Units are in MeV.

		density : float, optional
			Density of the element in g/cm^3.  Default behavior is to use
			`Element.density`.

		Other Parameters
		----------------
		**kwargs
			Optional keyword arguments for plotting.  See the 
			plotting section of the curie API for a complete
			list of kwargs.

		Examples
		--------
		>>> el = ci.Element('Ar')
		>>> el.plot_range()
		>>> el.plot_range(density=0.5)

		"""

		if energy is None:
			energy = 10.0**np.arange(-1.5, 2.8, 0.05)

		f,ax = _init_plot(**kwargs)
		ax.plot(energy, self.range(energy, particle=particle, density=density), label='Range ({0}, Z={1})'.format(self.name, self.Z))

		ax.set_xlabel('Incident Energy (MeV)')
		ax.set_ylabel('Range (cm)')
		ax.set_xscale('log')
		ax.set_yscale('log')
		ax.legend()

		return _draw_plot(f, ax, **kwargs)


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import json
import numpy as np
import pandas as pd
import datetime as dtm
import copy

from scipy.optimize import curve_fit
from scipy.special import erfc
from scipy.interpolate import interp1d

from .data import _get_connection
from .plotting import _init_plot, _draw_plot, colormap
from .calibration import Calibration
from .isotope import Isotope
from .compound import Compound

class Spectrum(object):
	"""Gamma-ray spectrum from High-Purity Germanium (HPGe) detectors

	Provides methods for reading, converting, and fitting gamma-ray spectra
	from HPGe data.
	
	Parameters
	----------
	filename : str
		File path to the gamma-ray spectrum.  Supported file types are Ortec .Spe
		and .Chn files.

	Other Parameters
	----------------
	cb : str or ci.Calibration
		Calibration to use, or path to a calibration .json file.

	isotopes : list
		List of gamma-decaying isotopes observed in the spectrum.

	fit_config : dict
		Dictionary of fit configuration parameters passed as keyword arguments to
		`self.fit_peaks()`. See that function for more details on these arguments.

	Attributes
	----------
	cb : ci.Calibration
		Gamma-ray energy, efficiency and resolution calibration.

	isotopes : list
		List of gamma-decaying isotopes observed in the spectrum.

	fit_config : dict
		Dictionary of fit configuration parameters passed as keyword arguments to
		`self.fit_peaks()`. See that function for more details on these arguments.

	filename : str
		File path to the gamma-ray spectrum.

	hist : np.ndarray
		Histogram of counts observed in the spectrum.

	start_time : datetime.datetime
		Date-time marking the start of the count.

	live_time : float
		Total time the spectrum was counting, in seconds, minus the dead time.

	real_time : float
		Total time the spectrum was counting, in seconds.

	peaks : pd.DataFrame
		Table of fits to the gamma-ray data.  Includes counts, isotopes, energies,
		intensities, efficiencies, calculated decays and decay-rates, chi^2, and other
		information.  The complete list of columns is 'filename', 'isotope', 'energy', 'counts',
		'unc_counts', 'intensity', 'unc_intensity', 'efficiency', 'unc_efficiency',
		'decays', 'unc_decays', 'decay_rate', 'unc_decay_rate', 'chi2', 'start_time',
		'live_time', and 'real_time'.

	Examples
	--------
	>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
	>>> sp.isotopes = ['Eu-152', '40K']
	>>> sp.cb = 'example_calib.json'
	>>> sp.fit_config = {'bg':'quadratic', 'xrays':False}
	>>> sp.saveas('eu_calib_7cm.Chn')
	>>> sp = ci.Spectrum('eu_calib_7cm.Chn')

	"""

	def __init__(self, filename=None, **kwargs):
		self.filename = filename
		if 'cb' in kwargs:
			if type(kwargs['cb']==str):
				self._cb = Calibration(kwargs['cb'])
			else:
				self._cb = copy.deepcopy(kwargs['cb'])
		else:
			self._cb = Calibration()

		self._fit_config = {'snip_adj':1.0, 'R':0.1, 'alpha':0.9,
							'step':0.00, 'bg':'snip', 'skew_fit':False,
							'step_fit':False, 'SNR_min':4.0, 'A_bound':1.0,
							'mu_bound':1.0, 'sig_bound':1.0, 'xrays':False,
							'pk_width':7.5, 'E_min':75.0, 'I_min':0.05,
							'dE_511':3.5, 'multi_max':8}
		if 'fit_config' in kwargs:
			self.fit_config = kwargs['fit_config']
		self._ortec_metadata = {}

		if filename is not None:
			if os.path.exists(filename):
				print('Reading Spectrum {}'.format(filename))
				if filename.endswith('.Spe'):
					self._read_Spe(filename)
				elif filename.endswith('.Chn'):
					self._read_Chn(filename)
				else:
					raise ValueError('File type not supported: {}'.format(filename))
				self._snip = self._snip_bg()
				self._snip_interp = interp1d(np.arange(len(self.hist)), self._snip, bounds_error=False, fill_value=0.0)
			else:
				raise ValueError('File does not exist: {}'.format(filename))

		if 'isotopes' in kwargs:
			self.isotopes = kwargs['isotopes']
		else:
			self.isotopes = []

		self._peaks = None
		self._fits = None
		self._geom_corr = 1.0
		self._atten_corr = None

		self._gmls = None


	def _read_Spe(self, filename):

		with open(filename) as f:
			ln = f.readline()
			while ln:
				if ln.startswith('$'):
					section = ln.strip()[1:-1]
					if section=='DATA':
						L = int(f.readline().strip().split()[1])+1
						self.hist = np.fromfile(f, dtype=np.int64, count=L, sep='\n')
					else:
						self._ortec_metadata[section] = []
				else:
					self._ortec_metadata[section].append(ln.strip())
				ln = f.readline()

		self.start_time = dtm.datetime.strptime(self._ortec_metadata['DATE_MEA'][0], '%m/%d/%Y %H:%M:%S')
		self.live_time, self.real_time = tuple(map(float, self._ortec_metadata['MEAS_TIM'][0].split()))

		self.cb.engcal = list(map(float, self._ortec_metadata['MCA_CAL'][-1].split(' ')[:-1]))

	def _read_Chn(self, filename):

		with open(filename, 'rb') as f:
			det_no = np.frombuffer(f.read(6), dtype='i2')[1]
			sts = np.frombuffer(f.read(2), dtype='S2')[0].decode('utf-8')

			self.real_time, self.live_time = tuple(map(float, 0.02*np.frombuffer(f.read(8), dtype='i4')))

			st = np.frombuffer(f.read(12), dtype='S12')[0].decode('utf-8')
			months = {'jan':'01','feb':'02','mar':'03','apr':'04',
						'may':'05','jun':'06','jul':'07','aug':'08',
						'sep':'09','oct':'10','nov':'11','dec':'12'}
			start_time = '{0}/{1}/{2} {3}:{4}:{5}'.format(months[st[2:5].lower()],st[:2],('20' if st[7]=='1' else '19')+st[5:7],st[8:10],st[10:],sts)
			self.start_time = dtm.datetime.strptime(start_time, '%m/%d/%Y %H:%M:%S')

			L = np.frombuffer(f.read(4), dtype='i2')[1]
			self.hist = np.asarray(np.frombuffer(f.read(4*L), dtype='i4'), dtype=np.int64)
			f.read(4)

			self.cb.engcal = np.frombuffer(f.read(12),dtype='f4').tolist()


			shape = np.frombuffer(f.read(12), dtype='f4').tolist()
			self._ortec_metadata['SHAPE_CAL'] = ['3', ' '.join(map(str, shape))]
			f.read(228)
			L = np.frombuffer(f.read(1), dtype='i1')[0]
			if L:
				det_desc = np.frombuffer(f.read(L), dtype='S{}'.format(L))[0].decode('utf-8')
				self._ortec_metadata['SPEC_REM'] = ['DET# '+str(det_no), 'DETDESC# '+det_desc, 'AP# Maestro Version 7.01']
			if L<63:
				f.read(63-L)
			L = np.frombuffer(f.read(1), dtype='i1')[0]
			if L:
				sample_desc = np.frombuffer(f.read(L),dtype='S{}'.format(L))[0].decode('utf-8')
				self._ortec_metadata['SPEC_ID'] = [sample_desc]


	@property
	def cb(self):
		return self._cb

	@cb.setter
	def cb(self, _cb):
		if type(_cb)==str:
			self._cb = Calibration(_cb)
		else:
			self._cb = copy.deepcopy(_cb)


	@property
	def fit_config(self):
		return self._fit_config

	@fit_config.setter
	def fit_config(self, _fit_config):
		self._peaks = None
		self._fits = None
		for nm in _fit_config:
			self._fit_config[nm] = _fit_config[nm]

	@property
	def isotopes(self):
		return self._isotopes

	@isotopes.setter
	def isotopes(self, _isotopes):
		self._gmls = None
		self._peaks = None
		self._fits = None
		self._isotopes = _isotopes
	

	def __str__(self):
		return self.filename

	def __add__(self, other):
		if not type(other)==Spectrum:
			raise ValueError('Cannot add spectrum to type {}'.format(type(other)))

		if self.start_time==other.start_time:
			alpha = np.sum(self.hist)/float(np.sum(other.hist))
			dead_time = alpha*(self.real_time-self.live_time)+(1.0-alpha)*(other.real_time-other.live_time)
			self.real_time = alpha*self.real_time+(1.0-alpha)*other.real_time
			self.live_time = self.real_time-dead_time
		else:
			self.real_time += other.real_time
			self.live_time += other.live_time
			

		if len(self.hist)==len(other.hist):
			if len(self.cb.engcal)==len(other.cb.engcal):
				if len(np.where(self.cb.engcal==other.cb.engcal)[0])==len(self.cb.engcal):
					self.hist += other.hist
					self._snip = self._snip_bg()
					self._snip_interp = interp1d(np.arange(len(self.hist)), self._snip, bounds_error=False, fill_value=0.0)
					return self

		other_bins = other.cb.eng(np.arange(-0.5, len(other.hist)+0.5, 1.0))
		dNdE = np.array(other.hist, dtype=np.float64)/(other_bins[1:]-other_bins[:-1])
		f = interp1d(other.cb.eng(np.arange(len(other.hist))), dNdE, bounds_error=False, fill_value=0.0)

		bins = self.cb.eng(np.arange(-0.5, len(self.hist)+0.5, 1.0))
		edges = np.append(bins, bins).reshape((2,len(bins))).T.flatten()[1:-1].reshape((len(bins)-1, 2)).T
		if np.__version__>='1.16.0':
			e_grid = np.linspace(edges[0], edges[1], num=10).T
		else:
			e_grid = np.array([np.linspace(edges[0][n], edges[1][n], num=10) for n in range(len(edges[0]))])
		N = np.asarray(np.trapz(f(e_grid), e_grid, axis=1), dtype=np.int64)
		self.hist += np.random.poisson(N)
		self._snip = self._snip_bg()
		self._snip_interp = interp1d(np.arange(len(self.hist)), self._snip, bounds_error=False, fill_value=0.0)

		return self
		
	def rebin(self, N_bins):
		"""Rebin the histogram

		Rebins the histogram to the closest power of 2 to
		the given `N_bins`.  Energy and resolution calibrations
		will be adjusted to match the new bin length. Note N_bins must
		be less than the current histogram length.

		Parameters
		----------
		N_bins : int
			Number of desired bins. Rounds to closest power of 2, e.g. 1000 rounds to 1024.

		Examples
		--------
		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> print(len(sp.hist))
		16384
		>>> sp.rebin(1000)
		>>> print(len(sp.hist))
		1024

		"""

		L = len(self.hist)
		if N_bins>L:
			raise ValueError('N_bins: {0} must not be greater than current value: {1}'.format(N_bins, L))

		r = int(round(L/float(N_bins)))
		self.hist = np.sum(self.hist.reshape((int(L/r), r)), axis=1)
		self._snip = self._snip_bg()
		self._snip_interp = interp1d(np.arange(len(self.hist)), self._snip, bounds_error=False, fill_value=0.0)

		ec = self.cb.engcal
		self.cb.engcal = [ec[0], ec[1]*r]+([ec[2]*r**2] if len(ec)==3 else [])

		rc = self.cb.rescal
		self.cb.rescal = ([rc[0]*np.sqrt(r)] if len(rc)==1 else [rc[0], rc[1]*r])
		
	def attenuation_correction(self, compounds, x=None, ad=None): # first compound is 'self'
		"""Efficiency correction for sample attenuation

		Corrects the efficiency used in calculating observed decays from each peak for
		sample attenuation.  A list of compounds (and their areal densities) is given,
		and a correction factor as a function of energy is determined using the
		photon attenuation data from XCOM.  The first compound in the list is assumed
		to be the radiogenic sample (the "self"), and the other compounds are presumed
		between the detector and sample.  If it is desired to neglect self-attenuation,
		specify the thickness of the first compound to be zero.  Either thickness in units
		of centimeters must be given for each compound, or the areal density in
		units of g/cm^2.  Note that calling `sp.attenuation_correction()` will modify
		the efficiencies/decays in the peak fits, but it also returns the correction
		factor function used to do so.

		Parameters
		----------
		compounds : list of str or ci.Compound
			List of compounds to correct for.  The first compound is assumed to be
			the radiogenic sample ("self"), the following compounds are assumed to
			be between the source and detector.  If str, the compounds must be either
			a natural element, or be included in `ci.COMPOUND_LIST`.  They can also
			be a ci.Compound object.

		x : list of float
			List of the thicknesses (in cm) of each compound to be corrected for.
			len(x) must be the same as len(compounds).
			If a density other than the density of the corresponding compound is
			desired, use the `ad` keyword.  Only one of `x` or `ad` should be given,
			but at least one is required.

		ad : list of float
			List of the areal densities (in g/cm^2) of each compound to be corrected
			for.  len(ad) must be the same as len(compounds).  Only one of `x` or `ad`
			should be given, but at least one is required.

		Returns
		-------
		correction_factor : scipy.interpolate.interp1d
			Interpolation function that gives the correction factor
			as a function of energy, in keV.  Correction factor should
			be multiplied with the efficiency.  

		Examples
		--------
		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> print(sp.attenuation_correction(['Fe', ci.Compound('H2O', density=1.0)], x=[0.1, 0.5])(100*np.arange(1, 10)))
		[0.79614452 0.88215107 0.90281046 0.91410313 0.92187566 0.92780998
		 0.93217524 0.9365405  0.93969564]
		>>> print(sp.attenuation_correction(['La', ci.Compound('Kapton', density=12.0)], ad=[0.1, 0.5])(100*np.arange(1, 10)))
		[0.82705455 0.92060463 0.93878407 0.94717435 0.95248909 0.95638036
		 0.95914978 0.9619192  0.96387558]

		"""

		cm_ls = [Compound(cm) if type(cm)==str else cm for cm in compounds]
		energy = np.unique(np.concatenate([cm.mass_coeff['energy'].to_numpy() for cm in cm_ls]))

		if ad is None:
			ad = [x[n]*cm.density for n,cm in enumerate(cm_ls)]
		ad = np.asarray(ad)

		mu_0 = cm_ls[0].mu(energy)
		atten = (1.0-np.exp(-ad[0]*mu_0))/(ad[0]*mu_0)

		for n,cm in enumerate(cm_ls[1:]):
			atten *= cm.attenuation(energy, ad[n+1], 1.0)

		atten_corr = interp1d(energy, atten, bounds_error=False, fill_value='extrapolate')
		self._atten_corr = atten_corr
		return atten_corr

		
	def geometry_correction(self, distance, r_det, thickness, sample_size, shape='circle', N=1000000):
		"""Efficiency correction for sample geometry

		Correction to the efficiency for non-point source geometries.
		The correction factor is a multiplier to the efficiency, that corrects
		for a sample that was efficiency calibrated using a point source, but
		that is not a point source itself.  Note that the input sizes do not need
		units, but they are assumed to all be in the same units system (e.g. they
		are all in cm, or inches).  Also, like the attenuation_correction, the
		correction factor is automatically applied to the peak data, but it
		is returned by the function as well.

		Parameters
		----------
		distance : float
			Distance from the front of the sample to the detector.
			Units must be consistent with other inputs.

		r_det : float
			Radius of the detector. Units must be consistent with
			other inputs.

		thickness : float
			Thickness of the sample, in the same units as the
			other inputs.

		sample_size : float
			Characteristic sample dimensions (in same units).  If
			shape is circle, this is sample radius.  If square, it
			is side length.  If rectangle, sample_size must be
			a length 2 tuple/array/list, corresponding to
			the x and y dimensions of the rectangle.

		shape : str, optional
			Shape of the sample. Options are 'circle' (default),
			'square' and 'rectangle'.

		N : int, optional
			Number of particles to use in Monte-Carlo solid angle
			calculation.  Default is 1E6.  If the sampling error
			in the correction factor is greater than 1 percent,
			a warning will be given to increase N.

		Returns
		-------
		correction_factor : float
			Geometry correction factor, to be multiplied with the efficiency.

		Examples
		--------
		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> print(sp.geometry_correction(distance=4, r_det=5, thickness=0.1, sample_size=2, shape='square'))
		0.9744182633801829
		>>> print(sp.geometry_correction(distance=30, r_det=5, thickness=10, sample_size=1))
		0.7586316490264302
		>>> print(sp.geometry_correction(distance=4, r_det=5, thickness=0.1, sample_size=(2,1.5), shape='rectangle'))
		0.9784496516955806

		"""

		N = int(N)
		distance = float(distance)
		r_det = float(r_det)

		if shape.lower()=='circle':
			r = np.sqrt(np.random.uniform(size=N))*float(sample_size)
			ang = np.pi*2.0*np.random.uniform(size=N)
			x0 = r*np.cos(ang)
			y0 = r*np.sin(ang)

		elif shape.lower()=='square':
			x0 = (np.random.uniform(size=N)-0.5)*float(sample_size)
			y0 = (np.random.uniform(size=N)-0.5)*float(sample_size)

		elif shape.lower()=='rectangle':
			x0 = (np.random.uniform(size=N)-0.5)*float(sample_size[0])
			y0 = (np.random.uniform(size=N)-0.5)*float(sample_size[1])

		tht = np.arccos(2.0*np.random.uniform(size=N)-1.0)
		phi = 2.0*np.pi*np.random.uniform(size=N)

		z = distance+np.random.uniform(size=N)*float(thickness)
		xf = x0 + z*np.tan(tht)*np.cos(phi)
		yf = y0 + z*np.tan(tht)*np.sin(phi)

		N_i = float(len(np.where(np.sqrt(xf**2+yf**2)<=r_det)[0]))
		if N_i<1E4:
			print('WARNING: Uncertainty in solid-angle {}% -- Increase N'.format(round(1E2*np.sqrt(N_i)/N_i, 1)))

		SA = 2.0*np.pi*N_i/float(N)
		corr = SA/(2.0*np.pi*(1.0-distance/np.sqrt(distance**2+r_det**2)))
		self._geom_corr = corr

		return corr



	def _snip_bg(self):
		adj = self.fit_config['snip_adj']

		x, dead = np.arange(len(self.hist)), int(7.5*adj*self.cb.res(len(self.hist)))
		V_i, L = np.log(np.log(np.sqrt(self.hist+1.0)+1.0)+1.0), len(x)-dead
		while self.hist[dead]==0:
			dead += 1
		
		for M in np.linspace(0, 7.5*adj, 10):
			l, h = np.array(x-M*self.cb.res(x), dtype=np.int32), np.array(x+M*self.cb.res(x), dtype=np.int32)
			V_i[dead:L] = np.minimum(V_i[dead:L], 0.5*(V_i[l[dead:L]]+V_i[h[dead:L]]))

		snip = (np.exp(np.exp(V_i)-1.0)-1.0)**2-1.0
		snip += adj*1.5*np.sqrt(snip+0.01)

		a1, a2 = 0.3/(adj*self.cb.res(0.1*L)), 0.3/(adj*self.cb.res(0.9*L))
		wt1, wt2 = x[::-1]/x[-1], x/x[-1]

		def exp_smooth(x, alpha):
			N = int(2.0/alpha)-1
			wts = np.array([alpha*(1.0-alpha)**abs(N-i) for i in range(2*N+1)])
			y = np.concatenate((np.ones(N)*x[0], x, np.ones(N)*x[-1]))
			return np.dot(wts/np.sum(wts), np.take(y, [np.arange(i,len(x)+i) for i in range(2*N+1)]))

		return wt1*exp_smooth(snip, a1) + wt2*exp_smooth(snip, a2)


	def _forward_fit(self, engcal=None):
		itps, gm = self._gammas()

		L = len(self.hist)
		chan = np.arange(L)
		W = self.fit_config['pk_width']
		R, alpha, step = self.fit_config['R'], self.fit_config['alpha'], self.fit_config['step']
		X = np.zeros((len(itps), L))

		for n,ip in enumerate(itps):
			idx = self.cb.map_channel(gm[n]['energy'], engcal)
			sig = self.cb.res(idx)
			A_norm = self.cb.eff(gm[n]['energy'])*gm[n]['intensity'].to_numpy()/sig
			whr = np.where(((idx+1.5*W*sig)<L)&((idx-1.5*W*sig)>0))[0]
			idx, sig, A_norm = idx[whr], sig[whr], A_norm[whr]

			l, h = np.array(idx-W*sig, dtype=np.int32), np.array(idx+W*sig, dtype=np.int32)
			for m,i in enumerate(idx):
				X[n][l[m]:h[m]] += self._peak(chan[l[m]:h[m]], A_norm[m], i, sig[m], R, alpha, step)

		Y = self.hist-self._snip
		return X, Y, np.dot(np.linalg.inv(np.dot(X, X.T)), np.dot(X, Y.T).T)

	def auto_calibrate(self, guess=None, peaks=None):
		"""Attempt to automatically adjust the energy calibration

		Uses a genetic forward fitting algorithm to attempt to automatically
		adjust the energy calibration in cases where the peak centroids are
		significantly mis-calibrated.  The algorithm requires a list of 
		isotopes in the spectrum, so self.isotopes cannot be None.  Also,
		the feature will only make small adjustments to the calibration,
		so if the current calibration is greater than 0.5 percent off, 
		either a guess or a list of peak locations must be given

		Parameters
		----------
		guess : array_like, optional
			Guess parameters for the energy calibration. Must be a 
			length 2 or 3 array.  If None, the energy calibration
			from self.cb.engcal will be used.

		peaks : array_like, optional
			List of peak locations.  Must be a 2-d array of format [[channel, energy], ...]

		Examples
		--------
		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> print(sp.cb.engcal)
		[3.3973e-01 1.8297e-01 5.5683e-09]
		>>> sp.cb.engcal = [0.3, 0.184]
		>>> sp.isotopes = ['152EU']
		>>> sp.plot()

		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> sp.cb.engcal = [0.3, 0.1835]
		>>> sp.isotopes = ['152EU']
		>>> sp.auto_calibrate()
		>>> print(sp.cb.engcal)
		[0.3        0.18281362 0.        ]
		>>> sp.plot()

		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> sp.cb.engcal = [0.3, 0.1]
		>>> sp.isotopes = ['152EU']
		>>> sp.auto_calibrate(peaks=[[664, 121.8]])
		>>> print(sp.cb.engcal)
		[0.         0.18289791 0.        ]
		>>> sp.plot()

		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> sp.cb.engcal = [0.3, 0.1]
		>>> sp.isotopes = ['152EU']
		>>> sp.auto_calibrate(guess=[0.3, 0.1835])
		>>> print(sp.cb.engcal)
		[0.3        0.18281398 0.        ]
		>>> sp.plot()


		"""

		from scipy.optimize import differential_evolution

		if guess is None:
			guess = list(self.cb.engcal)

		if peaks is not None:
			peaks = np.asarray(peaks, dtype=np.float64)
			if len(peaks.shape)==1:
				peaks = np.array([peaks.tolist()])
			if len(peaks)==1:
				guess = [0.0, peaks[0][1]/peaks[0][0]]
			else:
				N = 2 if len(peaks)<5 else 3
				M = np.column_stack([peaks[:,0]**m for m in range(N)])
				b = np.array([np.sum(peaks[:,0]**m*peaks[:,1]) for m in range(N)])
				M_inv = np.linalg.inv(np.dot(M.T, M))
				guess = np.dot(M_inv, b).tolist()

		guess = guess if len(guess)==3 else list(guess)+[0.0]
		
		def obj(*m):
			X, Y, B = self._forward_fit(engcal=[guess[0], m[0], guess[2]])
			return np.sum(np.sqrt(np.abs(Y-np.dot(B, X))))

		self.cb.engcal = [guess[0], differential_evolution(obj, [(0.995*guess[1], 1.005*guess[1])]).x[0], guess[2]]


	def _gammas(self, _force=False):
		if self._gmls is None or _force:
			itps, gm = [], []
			for i in self.isotopes:
				g = Isotope(i).gammas(I_lim=self.fit_config['I_min'], E_lim=self.fit_config['E_min'], dE_511=self.fit_config['dE_511'], xrays=self.fit_config['xrays'])
				g['intensity'] = g['intensity']*1E-2
				g['unc_intensity'] = g['unc_intensity']*1E-2
				if len(g):
					itps.append(i)
					gm.append(g)
			self._gmls = (itps, gm)
			return itps, gm
		else:
			return self._gmls

	def _chi2(self, fit, l ,h):
		non_zero = np.where(self.hist[l:h]>0)
		dof = float(len(non_zero[0])-len(fit)-1)
		if dof==0:
			return np.inf
		resid = self.hist[l:h][non_zero]-self._multiplet(np.arange(l,h)[non_zero], *fit)
		return np.sum(resid**2/self.hist[l:h][non_zero])/dof
		
	def _unc_calc(self, fn, p0, cov):
		var, eps = 0.0, 1E-8
		for n in range(len(p0)):
			for m in range(n, len(p0)):
				c_n, c_m = list(p0), list(p0)
				c_n[n], c_m[m] = c_n[n]+eps, c_m[m]+eps
				par_n = (fn(*c_n)-fn(*p0))/eps
				par_m = (fn(*c_m)-fn(*p0))/eps
				var += cov[n][m]*par_n*par_m*(2.0 if n!=m else 1.0)
		return var
	
	def _counts(self, fit, cov):
		cfg = self.fit_config
		L = 3+2*int(cfg['skew_fit'])+int(cfg['step_fit'])
		M = {'snip':0, 'constant':1, 'linear':2, 'quadratic':3}[cfg['bg'].lower()]
		N_cts, unc_N = [], []

		skew_fn = lambda A, R, alpha, sig: 2*A*R*alpha*sig*np.exp(-0.5/alpha**2)
		min_skew_fn = lambda A, sig: 2*A*cfg['R']*cfg['alpha']*sig*np.exp(-0.5/cfg['alpha']**2)
		pk_fn = lambda A, sig: 2.506628*A*sig

		for m in range(int((len(fit)-M)/L)):
			i = L*m+M
			p_i = [i, i+2]
			s_i = [i, i+3, i+4, i+2]
			if cfg['skew_fit']:
				N_cts.append(pk_fn(*fit[p_i])+skew_fn(*fit[s_i]))
			else:
				N_cts.append(pk_fn(*fit[p_i])+min_skew_fn(*fit[p_i]))
			if cov is not None:
				if not np.isinf(cov[i][i]):
					if cfg['skew_fit']:
						skew_unc = self._unc_calc(skew_fn, fit[s_i], cov[np.ix_(s_i, s_i)])
					else:
						skew_unc = self._unc_calc(min_skew_fn, fit[p_i], cov[np.ix_(p_i, p_i)])
					unc_N.append(np.sqrt(self._unc_calc(pk_fn, fit[p_i], cov[np.ix_(p_i, p_i)])+skew_unc))
				else:
					unc_N.append(np.inf)
			else:
				unc_N.append(np.inf)
		return np.array(N_cts), np.array(unc_N)
	
	def _decays(self, N, unc_N, df):
		if self._atten_corr is None:
			corr = self._geom_corr
		else:
			corr = self._geom_corr*self._atten_corr(df['energy'])
			
		D = N/(df['intensity']*self.cb.eff(df['energy'])*corr*(self.live_time/self.real_time))
		unc_D = D*np.sqrt((N/N**2)+(unc_N/N)**2+(self.cb.unc_eff(df['energy'])/self.cb.eff(df['energy']))**2+(df['unc_intensity']/df['intensity'])**2)

		A, unc_A = D/self.real_time, unc_D/self.real_time
		return D, unc_D, A, unc_A

	def _peak(self, x, A, mu, sig, R, alpha, step):
		r2 = 1.41421356237
		return A*np.exp(-0.5*((x-mu)/sig)**2)+R*A*np.exp((x-mu)/(alpha*sig))*erfc((x-mu)/(r2*sig)+1.0/(r2*alpha))+step*A*erfc((x-mu)/(r2*sig))
		
	def _multiplet(self, x, *args):
		bg_fit = self.fit_config['bg'].lower()
		if bg_fit=='snip':
			b, peak = 0, self._snip_interp(x)
		elif bg_fit=='constant':
			b, peak = 1, args[0]*np.ones(len(x))
		elif bg_fit=='linear':
			b, peak = 2, args[0]+args[1]*x
		elif bg_fit=='quadratic':
			b, peak = 3, args[0]+args[1]*x+args[2]*x**2

		R, alpha, step = self.fit_config['R'], self.fit_config['alpha'], self.fit_config['step']
		if self.fit_config['skew_fit']:
			if self.fit_config['step_fit']:
				for n in range(int((len(args)-b)/6)):
					peak += self._peak(x,*args[6*n+b:6*n+6+b])
			else:
				for n in range(int((len(args)-b)/5)):
					peak += self._peak(x,*(args[5*n+b:5*n+5+b]+(step,)))
		else:
			if self.fit_config['step_fit']:
				for n in range(int((len(args)-b)/4)):
					peak += self._peak(x,*(args[4*n+b:4*n+b+3]+(R, alpha)+(args[4*n+3+b])))
			else:
				for n in range(int((len(args)-b)/3)):
					peak += self._peak(x,*(args[3*n+b:3*n+b+3]+(R,alpha,step)))
		return peak

	def _get_p0(self, gammas=None):
		### gammas must at least have energy, intensity, unc_intensity (isotope optional)
		istp, gm = self._gammas(True)
		for n,i in enumerate(istp):
			gm[n]['isotope'] = i
		if gammas is not None:
			gm += [pd.DataFrame(gammas)]
			gm[-1]['intensity'] = gm[-1]['intensity']*1E-2
			gm[-1]['unc_intensity'] = gm[-1]['unc_intensity']*1E-2


		X, Y, B = self._forward_fit()
		L = len(self.hist)
		chan = np.arange(L)

		if not len(gm):
			return []

		df = pd.concat(gm, sort=True, ignore_index=True).sort_values(by=['energy']).reset_index(drop=True)
		df['idx'] = self.cb.map_channel(df['energy'])
		df['sig'] = self.cb.res(df['idx'])
		df['l'] = np.array(df['idx']-self.fit_config['pk_width']*df['sig'], dtype=np.int32)
		df['h'] = np.array(df['idx']+self.fit_config['pk_width']*df['sig'], dtype=np.int32)
		df = df[(df['l']>0)&(df['h']<L)].reset_index(drop=True)
		df['A'] = (self.hist-self._snip)[df['idx']]

		for n,i in enumerate(istp):
			df_sub = df[df['isotope']==i]
			df.loc[df_sub.index, 'A'] = B[n]*self.cb.eff(df_sub['energy'])*df_sub['intensity']/df_sub['sig']

		df['SNR'] = df['A']/np.sqrt(self._snip[df['idx']])
		df = df[(df['SNR']>self.fit_config['SNR_min'])].reset_index(drop=True)

		if len(df)==0:
			return []

		multiplets = [df.loc[[0]]]
		for n,l in enumerate(df['l'].to_numpy()):
			if n>0:
				if l<multiplets[-1].loc[n-1, 'h'] and len(multiplets[-1])<self.fit_config['multi_max']:
					multiplets[-1] = pd.concat([multiplets[-1], df.loc[[n]]])
				else:
					multiplets.append(df.loc[[n]])

		p0 = []
		for multi in multiplets:
			p = {'df':multi, 'l':multi['l'].min(), 'h':multi['h'].max(), 'p0':[], 'bounds':[[],[]], 'istp':multi['isotope'].to_list()}

			bgf = self.fit_config['bg'].lower()
			if bgf=='constant':
				p['p0'].append(np.average(self._snip[p['l']:p['h']]))
				p['bounds'][0].append(0.85*p['p0'][0])
				p['bounds'][1].append(1.15*p['p0'][0])

			elif bgf=='linear' or bgf=='quadratic':
				N = 2 if bgf=='linear' else 3
				x = chan[p['l']:p['h']]
				M = np.column_stack([x**m for m in range(N)])
				b = np.array([np.sum(x**m*self._snip[p['l']:p['h']]) for m in range(N)])
				M_inv = np.linalg.inv(np.dot(M.T, M))
				p['p0'] += np.dot(M_inv, b).tolist()
				resid = self._snip[p['l']:p['h']]-np.dot(M, p['p0'])
				unc = np.sqrt(np.abs(M_inv*np.dot(resid.T, resid)/max([float((p['h']-p['l'])-N), 1.0])))
				p['bounds'][0] += [i-100*unc[n][n] for n,i in enumerate(p['p0'])]
				p['bounds'][1] += [i+100*unc[n][n] for n,i in enumerate(p['p0'])]

			R, alpha, step = self.fit_config['R'], self.fit_config['alpha'], self.fit_config['step']
			bA, bm, bs = 10.0*self.fit_config['A_bound'], 1.5*self.fit_config['mu_bound'], 1.5*self.fit_config['sig_bound']
			for n,rw in multi.iterrows():
				p['p0'] += [rw['A'], rw['idx'], rw['sig']]
				p['bounds'][0] += [0.0, rw['idx']-bm*rw['sig'], rw['sig']/bs]
				p['bounds'][1] += [rw['A']*bA, rw['idx']+bm*rw['sig'], rw['sig']*bs]

				if self.fit_config['skew_fit']:
					p['p0'] += [R, alpha]
					p['bounds'][0] += [0.0,0.5]
					p['bounds'][1] += [1.0, max((2.5, alpha))]

				if self.fit_config['skew_fit']:
					p['p0'] += [step]
					p['bounds'][0] += [0.0]
					p['bounds'][1] += [0.1]

			p0.append(p)

		return p0

	def _split_fits(self):
		fits, istp, lh, N_sub = [], [], [], []
		cfg = self.fit_config
		if self._fits is None:
			self.fit_peaks()

		for ft in self._fits:
			f = ft['fit']
			B = {'snip':0, 'constant':1, 'linear':2, 'quadratic':3}[cfg['bg'].lower()]
			p0 = f[:B].tolist()
			L = 3+2*int(cfg['skew_fit'])+int(cfg['step_fit'])
			N_sub.append(int((len(f)-B)/L))
			for n in range(N_sub[-1]):
				if type(ft['istp'][n])==str:
					istp.append(ft['istp'][n])
				else:
					istp.append(None)
				fits.append(p0+f[B+n*L:B+(n+1)*L].tolist())
				mu, sig = f[B+n*L+1], f[B+n*L+2]
				lh.append([mu-cfg['pk_width']*sig, mu+cfg['pk_width']*sig])
		itp_set = sorted(list(set(istp)))
		return {i:[[fits[n],lh[n]] for n,ip in enumerate(istp) if ip==i] for i in itp_set}, itp_set, N_sub


	def _multi_fit(self, p0):
		chan = np.arange(len(self.hist))
		try:
			fit, unc = curve_fit(self._multiplet, chan[p0['l']:p0['h']], self.hist[p0['l']:p0['h']], p0=p0['p0'], bounds=p0['bounds'], sigma=np.sqrt(self.hist[p0['l']:p0['h']]+0.1))
			p0['fit'] = fit
			p0['unc'] = unc
			N, unc_N = self._counts(fit, unc)
			D, unc_D, A, unc_A = self._decays(N, unc_N, p0['df'])
			chi2 = self._chi2(fit, p0['l'], p0['h'])

			f = p0['df']
			cols = ['filename','isotope','energy','counts','unc_counts',
				'intensity','unc_intensity','efficiency','unc_efficiency',
				'decays','unc_decays','decay_rate','unc_decay_rate','chi2',
				'start_time', 'live_time', 'real_time']

			df = pd.DataFrame({'filename':self.filename, 'isotope':f['isotope'], 'energy':f['energy'],
							'counts':N, 'unc_counts':unc_N, 'intensity':f['intensity'],
							'unc_intensity':f['unc_intensity'], 'efficiency':self.cb.eff(f['energy']),
							'unc_efficiency':self.cb.unc_eff(f['energy']), 'decays':D,
							'unc_decays':unc_D, 'decay_rate':A, 'unc_decay_rate':unc_A, 'chi2':chi2,
							'start_time':dtm.datetime.strftime(self.start_time, '%m/%d/%Y %H:%M:%S'),
							'live_time':self.live_time, 'real_time':self.real_time}, columns=cols)

			return p0, df
		except:
			return p0, p0['df']

	def fit_peaks(self, gammas=None, **kwargs):
		"""Fit the peaks in the spectrum

		Fits a multiplet of peaks, with configurable parameters, to the spectrum.
		The list of gammas can either be generated automatically by setting the
		`Spectrum.isotopes` attribute, or by using the 'gammas' keyword, which
		adds to any gammas generated from the isotopes attribute.  This function is
		called once when any of the following happen: the `Spectrum.peaks` property is
		accessed, the `Spectrum.saveas()` method is called, or the `Spectrum.summarize()`
		or `Spectrum.plot()` methods are called.  The peaks are then saved, and won't be
		re-fit unless `Spectrum.fit_peaks()` (this method) is called explicitly.  This
		is important if parameters such as the calibration, isotopes, or fit_config are changed
		after the peaks are generated the first time.

		Parameters
		----------
		gammas : list, dict or pd.DataFrame, optional
			Manual entry for specifying peaks in the spectrum. 'gammas' must be an
			object that can be converted into a pandas DataFrame, with the keywords
			'energy', 'intensity', 'unc_intensity' and optionally 'isotope'.  Units
			of intensity should be percent, units of energy should be keV.

		Other Parameters
		----------------
		xrays : bool
			Whether peak fits should include x-rays, as given by Nudat2. Default False.

		E_min : float
			Minimum peak energy to fit, in keV. Default 75.0.

		I_min : float
			Minimum peak intensity to fit, in percent. Default 0.05

		dE_511 : float
			Gammas that are fewer than dE_511 keV from the 511 keV
			annihilation peak are excluded from the fit. Default 3.5

		bg : str
			Type of background fit to use. Options are 'constant', 'linear',
			'quadratic' or 'snip'. Default 'snip'.  The 'snip' algorithm interpolates 
			the background under the peaks using the detector resolution, and the
			assumption that the background is smoothly varying.  Will fail for
			peaks that are on top of quickly varying features such as the electron
			backscatter peak, but has the benefit of removing a free parameter from
			the fit.

		skew_fit : bool
			If skew_fit is True, the fit parameters include the skewed gaussian component,
			otherwise only the gaussian parameters are fit. Default False.  The skewed gaussian is 
			characterized by the parameters R and alpha, and has the functional
			form R*A*exp((x-mu)/(alpha*sig))*erfc((x-mu)/(sqrt(2)*sig)+1.0/(sqrt(2)*alpha))
			where A, mu and sigma are the amplitude, centroid and width of the gaussian
			peak function.  Note that unless R and alpha are explicitly set to zero, they
			will be included in the peak function.  skew_fit only specifies whether or not
			their values are to be fit for each peak.

		step_fit : bool
			If step_fit is True, a step function is added to the background fit,
			which arises from compton scattering. Default False. The functional
			form of the step function is step*A*erfc((x-mu)/(sqrt(2)*sig)), where
			A, mu and sigma are the amplitude, centroid and width of the gaussian peak.
			Similar to skew_fit, unless step is explicitly set to zero (it is by default),
			a step component is included in every peak.  step_fit only specifies whether or
			not the value of this parameter is fit for every peak.

		R : float
			Expected value of R (amplitude of skewed component of gaussian). Default 0.1.

		alpha : float
			Expected value of alpha (width of skewed component of gaussian). Default 0.9.

		step : float
			Expected value of step (amplitude of compton background). Default 0.0.

		pk_width : float
			Number of peak standard deviations (width) to include in the spectrum data passed to
			the curve fit function. Default 7.5.  Should be wide enough such that several background
			channels not in the peak can be included in the fit.  If too wide, multiplets may
			begin to overlap.

		snip_adj : float
			Multiplier for parameters in the snip background calculation. Default 1.0.

		SNR_min : float
			Minimum acceptable peak amplitude to noise (sqrt(background)) ratio. Default 4.0.

		A_bound : float
			Multiplier for the bounds on the fit parameter A (peak amplitude). Default 1.0.

		mu_bound : float
			Multiplier for the bounds on the fit parameter mu (peak centroid). Default 1.0.

		sig_bound : float
			Multiplier for the bounds on the fit parameter sig (peak width). Default 1.0.

		multi_max : int
			Maximum number of peaks allowed in a multiplet. Default 8.

		Returns
		-------
		peaks : pd.DataFrame
			Table of peaks that were successfully fit.

		Examples
		--------
		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> sp.isotopes = ['152EU']
		>>> print(sp.fit_peaks(SNR_min=5, dE_511=12))
		>>> print(sp.fit_peaks(bg='quadratic'))

		"""

		self.fit_config = kwargs
		p0 = self._get_p0(gammas)
		if len(p0):
			multiplets = list(map(self._multi_fit, p0))
			self._fits = [i[0] for i in multiplets if 'fit' in i[0]]
			self._peaks = pd.concat([i[1] for i in multiplets if 'fit' in i[0]], ignore_index=True)
		else:
			self._fits = []
			self._peaks = None

		return self._peaks


	@property
	def peaks(self):
		if self._peaks is None:
			self._peaks = self.fit_peaks()
		return self._peaks
	
		
	def saveas(self, filename, replace=False):
		"""Save the spectrum or peak information to a file

		If the file type is one of '.png', '.pdf', '.svg' or another graphical
		file type, a plot of the spectrum will be saved.  If it is one of the supported
		spectra formats, '.Spe', '.Chn' or '.spe', the spectrum will be converted and saved.  Note
		that only '.Spe' and '.Chn' can be read by ci.Spectrum.  If the file type is one of
		'.csv', '.db' or '.json', the peak data will be saved.  All three can be read by the
		ci.DecayChain.get_counts method.

		Parameters
		----------
		filename : str
			Name of the file to save to. Supported file types are '.png', '.pdf', '.eps', '.pgf', '.ps',
			'.raw', '.rgba', '.svg', '.svgz', '.Spe', '.Chn', '.spe', '.csv', '.json', and '.db'.

		replace : bool, optional
			If True, when saving to one of the three supported peak data file types, any existing file
			will be completely overwritten. If False, then only peak data that matches the spectrum
			name will be overwritten. Default, False.

		Examples
		--------
		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> sp.isotopes = ['152EU']
		>>> sp.saveas('test_plot.png')
		>>> sp.saveas('eu_calib.Chn')
		>>> sp.saveas('peak_data.csv')

		"""

		if any([filename.endswith(e) for e in ['.png','.pdf','.eps','.pgf','.ps','.raw','.rgba','.svg','.svgz']]):
				self.plot(saveas=filename, show=False)

		if filename.endswith('.Spe'):
			### Maestro ASCII .Spe ###
			self._ortec_metadata['DATE_MEA'] = [dtm.datetime.strftime(self.start_time, '%m/%d/%Y %H:%M:%S')]
			self._ortec_metadata['MEAS_TIM'] = ['{0} {1}'.format(int(self.live_time), int(self.real_time))]

			self._ortec_metadata['ENER_FIT'] = ['{0} {1}'.format(self.cb.engcal[0], self.cb.engcal[1])]
			self._ortec_metadata['MCA_CAL'] = ['3','{0} {1} {2} keV'.format(self.cb.engcal[0], self.cb.engcal[1], (self.cb.engcal[2] if len(self.cb.engcal)>2 else 0.0))]

			defaults = {'ROI':['0'],'SPEC_REM':['DET# 0','DETDESC# None','AP# Maestro Version 7.01'],'PRESETS':['0'],
						'SHAPE_CAL':['3','0E+00 0E+00 0E+00'],'SPEC_ID':['No sample description was entered.']}

			for d in defaults:
				if d not in self._ortec_metadata:
					self._ortec_metadata[d] = defaults[d]

			ss = '\n'.join(['${}:\n'.format(sc)+'\n'.join(self._ortec_metadata[sc]) for sc in ['SPEC_ID','SPEC_REM','DATE_MEA','MEAS_TIM']])
			ss += '\n$DATA:\n0 {}\n'.format(len(self.hist)-1)
			ss += '\n'.join([' '*(8-len(i))+i for i in map(str, self.hist)])+'\n'
			ss += '\n'.join(['${}:\n'.format(sc)+'\n'.join(self._ortec_metadata[sc]) for sc in ['ROI','PRESETS','ENER_FIT','MCA_CAL','SHAPE_CAL']])+'\n'

			with open(filename,'w') as f:
				f.write(ss)

		if filename.endswith('.Chn'):
			### Maestro integer .Chn ###
			if 'SPEC_REM' in self._ortec_metadata:
				det_no = int(self._ortec_metadata['SPEC_REM'][0].split(' ')[1].strip())
			else:
				det_no = 0

			ss = np.array([-1, det_no, 1], dtype='i2').tobytes()

			st = self.start_time
			ss += np.array(('0' if st.second<10 else '')+str(st.second), dtype='S2').tobytes()
			ss += np.array([self.real_time/0.02, self.live_time/0.02], dtype='i4').tobytes()
			months = {1:'Jan',2:'Feb',3:'Mar',4:'Apr',5:'May',6:'Jun',7:'Jul',8:'Aug',9:'Sep',10:'Oct',11:'Nov',12:'Dec'}
			ss += np.array(('0' if st.day<10 else '')+str(st.day)+months[st.month]+str(st.year)[-2:]+('1' if st.year>1999 else '0'), dtype='S8').tobytes()
			ss += np.array(('0' if st.hour<10 else '')+str(st.hour)+('0' if st.minute<10 else '')+str(st.minute), dtype='S4').tobytes()

			ss += np.array([0, len(self.hist)], dtype='i2').tobytes()
			ss += np.array(self.hist, dtype='i4').tobytes()
			ss += np.array([-102, 0], dtype='i2').tobytes()

			ss += np.array(self.cb.engcal, dtype='f4').tobytes()

			if 'SHAPE_CAL' in self._ortec_metadata:
				ss += np.array(self._ortec_metadata['SHAPE_CAL'][1].split(' '), dtype='f4').tobytes()
			else:
				ss += np.array([0,0,0], dtype='f4').tobytes()
			ss += np.zeros(228, dtype='i1').tobytes()

			if 'SPEC_REM' in self._ortec_metadata:
				L = min((63, len(self._ortec_metadata['SPEC_REM'][1].split('# ')[1])))
				ss += np.array(L, dtype='i1').tobytes()
				ss += np.array(self._ortec_metadata['SPEC_REM'][1].split('# ')[1][:L], dtype='S{}'.format(L)).tobytes()
			else:
				ss += np.array(0, dtype='i1').tobytes()

			if 'SPEC_ID' in self._ortec_metadata:
				if self._ortec_metadata['SPEC_ID'][0]!='No sample description was entered.':
					L = len(''.join(self._ortec_metadata['SPEC_ID']))
					L = min((63, L))
					ss += np.array(L, dtype='i1').tobytes()
					ss += np.array(''.join(self._ortec_metadata['SPEC_ID'])[:L])
				else:
					ss += np.array(0, dtype='i1').tobytes()
					ss += np.array('\x00'*63, dtype='S63').tobytes()
			else:
				ss += np.array(0, dtype='i1').tobytes()
				ss += np.array('\x00'*63, dtype='S63').tobytes()

			ss += np.array('\x00'*128, dtype='S128').tobytes()
			with open(filename, 'wb') as f:
				f.write(ss)


		if filename.endswith('.spe'):
			### Radware gf3 .spe ###
			name = self.filename[:8]+' '*(8-len(self.filename))
			with open(filename, 'wb') as f:
				ss = np.array(24, dtype=np.uint32).tobytes()
				ss += np.array(name, dtype='c').tobytes()
				ss += np.array([len(self.hist), 1, 1, 1, 24, 4*len(self.hist)], dtype=np.uint32).tobytes()
				ss += np.array(self.hist, dtype=np.float32).tobytes()
				ss += np.array(4*len(self.hist), dtype=np.uint32).tobytes()
				f.write(ss)

		if filename.endswith('.csv'):
			if os.path.exists(filename) and not replace:
				df = pd.read_csv(filename, header=0)
				df = df[df['filename']!=self.filename]
				df = pd.concat([df, self.peaks])
				df.to_csv(filename, index=False)
			else:
				self.peaks.to_csv(filename, index=False)

		if filename.endswith('.db'):
			if os.path.exists(filename) and not replace:
				con = _get_connection(filename)
				df = pd.read_sql('SELECT * FROM peaks', con)
				df = df[df['filename']!=self.filename]
				df = pd.concat([df, self.peaks])
				df.to_sql('peaks', con, if_exists='replace', index=False)
			else:
				self.peaks.to_sql('peaks', _get_connection(filename), if_exists='replace', index=False)

		if filename.endswith('.json'):
			if os.path.exists(filename) and not replace:
				df = pd.read_json(filename, orient='records')
				df = df[df['filename']!=self.filename][self.peaks.columns]
				df = pd.concat([df, self.peaks])
				json.dump(json.loads(df.to_json(orient='records')), open(filename, 'w'), indent=4)
			else:
				json.dump(json.loads(self.peaks.to_json(orient='records')), open(filename, 'w'), indent=4)

		
	def summarize(self):
		"""Summarize the fitted peaks in the spectrum

		Prints a summary of the observed counts in each peak, the
		number of decays of the given isotope it corresponds to, and an 
		estimate of the activity of the isotope.

		Examples
		--------
		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> sp.isotopes = ['152EU']
		>>> sp.summarize()
		152EU - 121.7817 keV (I = 28.53%)
		---------------------------------
		counts: 498785 +/- 3195
		decays: 2.515e+07 +/- 2.702e+06
		activity (Bq): 1.025e+04 +/- 1.101e+03
		activity (uCi): 2.770e-01 +/- 2.976e-02
		chi2/dof: 21.923
		...

		"""

		for n,p in self.peaks.iterrows():
			ln1 = [p['energy'], p['isotope'], p['intensity']*100.0]
			ln0 = '{1} - {0} keV (I = {2}%)'.format(*ln1)
			ss = ln0 + '\n'
			ss += ''.join(['-']*len(ln0)) + '\n'
			ln2 = [int(p['counts']), (int(p['unc_counts']) if np.isfinite(p['unc_counts']) else np.inf),
					format(p['decays'], '.3e'), format(p['unc_decays'], '.3e'), 
					format(p['decay_rate'], '.3e'), format(p['unc_decay_rate'], '.3e'), 
					format(p['decay_rate']/3.7E4, '.3e'), format(p['unc_decay_rate']/3.7E4, '.3e'), 
					round(p['chi2'], 3)]
			ss += 'counts: {0} +/- {1}'.format(ln2[0], ln2[1]) + '\n'
			ss += 'decays: {0} +/- {1}'.format(ln2[2], ln2[3]) + '\n'
			ss += 'activity (Bq): {0} +/- {1}'.format(ln2[4], ln2[5]) + '\n'
			ss += 'activity (uCi): {0} +/- {1}'.format(ln2[6], ln2[7]) + '\n'
			ss += 'chi2/dof: {}'.format(ln2[8]) + '\n'
			print(ss+'\n')

		
	def plot(self, fit=True, xcalib=True, **kwargs):
		"""Plot the spectrum

		Draws a plot of the spectrum, and any successful peak fits.  Peaks
		are colored by isotope, and a dashed grey line is drawn over multiplets
		to help evaluate goodness of fit.

		Parameters
		----------
		fit : bool, optional
			If True, include peak fits. Else, only the spectrum is drawn. Default, True.

		xcalib : bool, optional
			If True, the x-axis is the calibrated energy in keV. If False, it is the 
			ADC channel number.  This may help locate peaks to give the 
			`Spectrum.auto_calibrate()` function if the energy calibration is poor.
			Default, True.

		Other Parameters
		----------------
		**kwargs
			Optional keyword arguments for plotting.  See the 
			plotting section of the curie API for a complete
			list of kwargs.

		Examples
		--------
		>>> sp = ci.Spectrum('eu_calib_7cm.Spe')
		>>> sp.isotopes = ['152EU']
		>>> sp.plot()
		>>> sp.plot(xcalib=False)
		>>> sp.plot(style='poster')

		"""

		f, ax = _init_plot(figsize=(12.8, 4.8), **kwargs)

		chan = np.arange(len(self.hist))
		xgrid = np.array([chan-0.5, chan+0.5]).T.flatten()
		spec = np.array([self.hist, self.hist]).T.flatten()
		erange = self.cb.eng(xgrid) if xcalib else xgrid

		spec_label = self.filename.split('/')[-1] if self.filename is not None else None
		ax.plot(erange, spec, lw=1.2, zorder=1, label=spec_label)

		lbs = []
		if fit:
			# if self.peaks is not None:
			# 	if len(self.peaks)>0:
					
			cm, cl = colormap(), colormap(aslist=True)
			sub, itp, N_sub = self._split_fits()
			
			for n,p in enumerate(self._fits):
				if N_sub[n]>1:
					xgrid = np.arange(p['l'], p['h'], 0.1)
					pk_fit = self._multiplet(xgrid, *p['fit'])
					erange = self.cb.eng(xgrid) if xcalib else xgrid
					ax.plot(erange, np.where(pk_fit>0.1, pk_fit, 0.1), ls='--', lw=1.4, color=cm['gy'], alpha=0.8)

			
			cl = [c for c in cl if c not in [cm['k'], cm['gy']]]
			ls = ['-','-.','--']

			for n,i in enumerate(itp):
				c = cl[n%len(cl)]
				ilbl = Isotope(i).TeX if type(i)==str else None

				for p,lh in sub[i]:
					xgrid = np.arange(lh[0], lh[1], 0.1)
					erange = self.cb.eng(xgrid) if xcalib else xgrid
					pk_fit = self._multiplet(xgrid, *p)

					lb = ilbl if (ilbl not in lbs and len(lbs)<20) else None
					if lb is not None:
						lbs.append(lb)

					ax.plot(erange, np.where(pk_fit>0.1, pk_fit, 0.1), lw=1.4, color=c, label=lb, ls=ls[int(n/len(cl))%len(ls)])
							
		
		if xcalib:
			ax.set_xlabel('Energy (keV)')
		else:
			ax.set_xlabel('ADC Channel')
		ax.set_ylabel('Counts')
		
		ax.legend(loc=0, ncol=min((max((int(len(lbs)/5), 1)), 3)))
		kwargs['_default_log'] = True

		return _draw_plot(f, ax, **kwargs)
		


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re
import numpy as np
import pandas as pd

from .data import _get_connection
from .isotope import Isotope

class Library(object):
	"""Library of nuclear reaction data

	Provides a means of searching and retrieving data from various nuclear reaction
	libraries.  The currently available libraries are ENDF/B-VII.1, TENDL-2015, IRDFF-II,
	and the IAEA Medical Monitor reaction library.  For neutrons, the libraries are
	categorized by either the exlusive reaction, as in ENDF, TENDL, and partially IRDFF,
	or by the residual product (rp), in TENDL and partially IRDFF.  For charged particles,
	all libraries (TENDL, IAEA) are categorized by the residual product.  As such, TENDL
	is split into separate libraries based on incident particle, and by exclusive reaction
	type or residual product.
	
	Parameters
	----------
	name : str
		The name of the library.  For exclusive neutron reactions, use 'endf', 'tendl', 
		or 'irdff'.  For residual product neutron reactions, use 'tendl_n' or 'irdff'.  Use
		'tendl_p' for proton reactions, 'tendl_d' for deuteron reactions, or 'iaea' for either.

	Examples
	--------
	>>> lb = ci.Library('tendl_n')
	>>> print(lb.name)
	TENDL-2015
	>>> lb = ci.Library('endf')
	>>> print(lb.name)
	ENDF/B-VII.1

	"""

	def __init__(self, name):
		name = name.lower()
		if name in ['endf']:
			self.db_name = 'endf'
		elif name in ['tendl']:
			self.db_name = 'tendl'
		elif name in ['tendl_n_rp','tendl_nrp','tendl_n','nrp','rpn']:
			self.db_name = 'tendl_n_rp'
		elif name in ['tendl_p_rp','tendl_prp','tendl_p','prp','rpp']:
			self.db_name = 'tendl_p_rp'
		elif name in ['tendl_d_rp','tendl_drp','tendl_d','drp','rpd']:
			self.db_name = 'tendl_d_rp'
		elif name in ['irdff']:
			self.db_name = 'irdff'
		elif name in ['iaea','iaea-cpr','iaea-monitor','cpr','iaea_cpr','iaea_monitor','medical','iaea-medical','iaea_medical']:
			self.db_name = 'iaea_medical'
		else:
			raise ValueError('Library {} not recognized.'.format(name))

		self._con = _get_connection(self.db_name)
		self.name = {'endf':'ENDF/B-VII.1','tendl':'TENDL-2015','irdff':'IRDFF-II','iaea':'IAEA CP-Reference (2017)'}[self.db_name.split('_')[0]]
		self._warn = True

	def __str__(self):
		return self.name
		

	def search(self, target=None, incident=None, outgoing=None, product=None, _label=False):
		"""Searches the library for reactions

		Returns a list of reactions that match the specified target, incident
		or outgoing projectile, or product.  Any combination of arguments can
		be specified, but at least one must be specified.  Note that if the library
		is a TENDL residual-product library, outgoing does not need to be specified.

		Parameters
		----------
		target : str, optional
			The target nucleus.  Some libraries support natural elements, e.g. 'natEl'.

		incident : str, optional
			Incident particle.  Must be one of 'n', 'p' or 'd'.  Only needed for IAEA
			library with multiple incident projectiles optional.

		outgoing : str, optional
			Outgoing particle, or reaction shorthand.  E.g. '2n', 'd', 'f', 'inl', 'x'.
			Does not need to be specified for TENDL residual product libraries.

		product : str, optional
			The product isotope.  Not required for some exclusive reactions, e.g. 
			'115IN(n,g)116IN' is the same as '115IN(n,g)', but not the same as
			'115IN(n,g)116INm1'.

		Returns
		-------
		available_reactions : list of str (reaction names)
			Reactions that were found in the libary matching the search criteria.

		Examples
		--------
		>>> lb = ci.Library('tendl_p')
		>>> print(lb.search(target='Sr-86', product='Y-86g'))
		['86SR(p,x)86Yg']
		>>> lb = ci.Library('endf')
		>>> print(lb.search(target='226RA', product='225RA'))
		['226RA(n,2n)225RA']

		"""

		if incident is not None:
			incident = incident.lower()

		if outgoing is not None:
			outgoing = outgoing.lower()

		if target is not None:
			if '-' in target:
				if target.endswith('m') or target.endswith('g'):
					target = Isotope(target).name
				else:
					target = Isotope(target)._short_name

		if product is not None:
			if '-' in product:
				if product.endswith('m') or product.endswith('g'):
					product = Isotope(product).name
				else:
					product = Isotope(product)._short_name

		ss = 'SELECT * FROM all_reactions'

		if self.db_name in ['endf','tendl','irdff']:
			if incident is not None:
				if incident!='n':
					return []
			q = [('%'+i+'%' if not n%2 else i) for n,i in enumerate([target, outgoing, product]) if i]
			ss += ' WHERE ' if len(q) else ''
			ss += ' AND '.join([i for i in [('target LIKE ?' if target else ''),('outgoing=?' if outgoing else ''),('product LIKE ?' if product else '')] if i])
			reacs = [r.to_list() for n,r in pd.read_sql(ss, self._con, params=tuple(q)).iterrows()]
			fmt = '{0}(n,{1}){2}'

		elif self.db_name in ['tendl_n_rp','tendl_p_rp','tendl_d_rp']:
			if incident is not None:
				if incident!=self.db_name.split('_')[1]:
					return []

			if product:
				if 'm' not in product and 'g' not in product:
					product += 'g'
					if self._warn:
						print('WARNING: Product isomeric state not specified, ground state assumed.')
						self._warn = False

			q = ['%'+i+'%' for i in [target, product] if i]
			ss += ' WHERE ' if len(q) else ''
			ss += ' AND '.join([i for i in [('target LIKE ?' if target else ''),('product LIKE ?' if product else '')] if i])
			reacs = [r.to_list() for n,r in pd.read_sql(ss, self._con, params=tuple(q)).iterrows()]
			fmt = '{0}('+self.db_name.split('_')[1]+',x){1}'

		elif self.db_name=='iaea_medical':
			q = [('%'+i+'%' if n in [0,3] else i) for n,i in enumerate([target, incident, outgoing, product]) if i]
			ss += ' WHERE ' if len(q) else ''
			ss += ' AND '.join([i for i in [('target LIKE ?' if target else ''),('incident=?' if incident else ''),('outgoing=?' if outgoing else ''),('product LIKE ?' if product else '')] if i])
			reacs = [r.to_list() for n,r in pd.read_sql(ss, self._con, params=tuple(q)).iterrows()]
			fmt = '{0}({1},{2}){3}'

		if target:
			reacs = [i for i in reacs if i[0].lower()==target.lower()]
		if _label:
			return [i[-1] for i in reacs]

		return [fmt.format(*i) for i in reacs]

		
	def retrieve(self, target=None, incident=None, outgoing=None, product=None):
		"""Pull reaction data from the library

		Returns a numpy ndarray containing the reaction energy grid (in MeV),
		the cross section (in mb), and for IRDFF and IAEA, the uncertainty in the 
		cross section (mb).  The arguments are the same as for `Library.search()`.

		Parameters
		----------
		target : str, optional
			The target nucleus.  Some libraries support natural elements, e.g. 'natEl'.

		incident : str, optional
			Incident particle.  Must be one of 'n', 'p' or 'd'.  Only needed for IAEA
			library with multiple incident projectiles optional.

		outgoing : str, optional
			Outgoing particle, or reaction shorthand.  E.g. '2n', 'd', 'f', 'inl', 'x'.
			Does not need to be specified for TENDL residual product libraries.

		product : str, optional
			The product isotope.  Not required for some exclusive reactions, e.g. 
			'115IN(n,g)116IN' is the same as '115IN(n,g)', but not the same as
			'115IN(n,g)116INm1'.

		Returns
		-------
		reaction : np.ndarray
			Numpy ndarray containing the reaction data. Energy is in MeV, cross section
			is in mb.  The first column is energy, the second column is cross section, and
			if the library provides uncertainties in cross section this will be in the third column.

		Examples
		--------
		>>> lb = ci.Library('endf')
		>>> print(lb.retrieve(target='226RA', product='225RA')[-8:])
		[[18.5      11.97908 ]
		 [18.75     10.30011 ]
		 [19.        8.62115 ]
		 [19.25      6.942188]
		 [19.5       5.263225]
		 [19.75      3.584263]
		 [19.875     2.744781]
		 [20.        1.9053  ]]

		"""

		labels = self.search(target, incident, outgoing, product, _label=True)

		if not len(labels)==1:
			raise ValueError('{0}({1},{2}){3}'.format(target, incident, outgoing, product)+' is not a unique reaction.')

		if not target:
			raise ValueError('Target Must be specified.')

		if '-' in target:
			if target.endswith('m') or target.endswith('g'):
				target = Isotope(target).name
			else:
				target = Isotope(target)._short_name

		if self.db_name in ['endf','tendl','tendl_n_rp','tendl_p_rp','tendl_d_rp']:
			table = ''.join(re.findall('[A-Z]+', target))+'_'+''.join(re.findall('[0-9]+', target))+('m' if 'm' in target else '')

			return pd.read_sql('SELECT energy,{0} FROM {1}'.format(labels[0], table), self._con).to_numpy()*(np.array([1E-6, 1E3]) if self.db_name=='endf' else np.ones(2))

		elif self.db_name=='irdff':
			return pd.read_sql('SELECT * FROM {}'.format(labels[0]), self._con).to_numpy()*np.array([1E-6, 1E3, 1E3])

		elif self.db_name=='iaea_medical':
			if incident is None:
				raise ValueError('Incident particle must be specified.')
			table = {'n':'neutron','p':'proton','d':'deuteron','h':'helion','a':'alpha','g':'gamma'}[incident]
			return pd.read_sql('SELECT energy,cross_section,unc_cross_section FROM {0} WHERE target LIKE ? AND product=?'.format(table), self._con, params=('%'+target+'%', labels[0])).to_numpy()
		


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import matplotlib.pyplot as plt

LIGHT = """#1abc9c #2ecc71 #3498db #9b59b6 #34495e #f1c40f #e67e22 #e74c3c #ecf0f1 #95a5a6
#81ecec #55efc4 #74b9ff #a29bfe #636e72 #ffeaa7 #fab1a0 #ff7675 #dfe6e9 #b2bec3
#7ed6df #badc58 #686de0 #e056fd #30336b #f6e58d #ffbe76 #ff7979 #dff9fb #95afc0
#00a8ff #4cd137 #273c75 #9c88ff #353b48 #fbc531 #e67e22 #e84118 #f5f6fa #7f8fa6
#00d2d3 #1dd1a1 #54a0ff #5f27cd #576574 #feca57 #e67e22 #ff6b6b #ecf0f1 #8395a7
#70a1ff #7bed9f #70a1ff #5352ed #57606f #eccc68 #ff7f50 #ff6b81 #ffffff #dfe4ea
#12CBC4 #C4E538 #0652DD #FDA7DF #833471 #FFC312 #EE5A24 #ED4C67 #ecf0f1 #95a5a6
#38ada9 #b8e994 #6a89cc #6a89cc #1e3799 #fad390 #fa983a #f8c291 #ecf0f1 #95a5a6
#2bcbba #26de81 #4b7bec #a55eea #778ca3 #fed330 #fd9644 #fc5c65 #ecf0f1 #d1d8e0
#9AECDB #55E6C1 #25CCF7 #D6A2E8 #3B3B98 #F8EFBA #FEA47F #FD7272 #ecf0f1 #CAD3C8
#63cdda #2ecc71 #778beb #cf6a87 #596275 #f7d794 #f3a683 #e77f67 #ecf0f1 #95a5a6
#34ace0 #33d9b2 #706fd3 #40407a #34495e #ffb142 #ff793f #ff5252 #f7f1e3 #d1ccc0
#34e7e4 #0be881 #4bcffa #575fcf #485460 #ffdd59 #ffc048 #ff5e57 #ecf0f1 #d2dae2
#7efff5 #32ff7e #18dcff #7d5fff #4b4b4b #fffa65 #ffaf40 #ff4d4d #ecf0f1 #95a5a6"""

DARK = """#16a085 #27ae60 #2980b9 #8e44ad #2c3e50 #f39c12 #d35400 #c0392b #bdc3c7 #7f8c8d
#00cec9 #00b894 #0984e3 #6c5ce7 #2d3436 #fdcb6e #e17055 #d63031 #b2bec3 #636e72
#22a6b3 #6ab04c #4834d4 #be2edd #130f40 #f9ca24 #f0932b #eb4d4b #c7ecee #535c68
#0097e6 #44bd32 #192a56 #8c7ae6 #2f3640 #e1b12c #d35400 #c23616 #dcdde1 #718093
#01a3a4 #10ac84 #2e86de #341f97 #222f3e #ff9f43 #d35400 #ee5253 #bdc3c7 #8395a7
#1e90ff #2ed573 #1e90ff #3742fa #2f3542 #ffa502 #ff6348 #ff4757 #f1f2f6 #ced6e0
#1289A7 #A3CB38 #1B1464 #D980FA #6F1E51 #F79F1F #EA2027 #B53471 #bdc3c7 #7f8c8d
#079992 #78e08f #4a69bd #4a69bd #0c2461 #f6b93b #e58e26 #e55039 #bdc3c7 #7f8c8d
#0fb9b1 #20bf6b #3867d6 #8854d0 #4b6584 #f7b731 #fa8231 #eb3b5a #bdc3c7 #a5b1c2
#BDC581 #58B19F #1B9CFC #82589F #182C61 #EAB543 #F97F51 #FC427B #bdc3c7 #2C3A47
#3dc1d3 #27ae60 #546de5 #c44569 #303952 #f5cd79 #f19066 #e15f41 #bdc3c7 #7f8c8d
#227093 #218c74 #474787 #2c2c54 #2c3e50 #cc8e35 #cd6133 #b33939 #aaa69d #84817a
#00d8d6 #05c46b #0fbcf9 #3c40c6 #1e272e #ffd32a #ffa801 #ff3f34 #bdc3c7 #808e9b
#67e6dc #3ae374 #17c0eb #7158e2 #3d3d3d #fff200 #ff9f1a #ff3838 #bdc3c7 #7f8c8d"""

def colormap(palette='default', shade='dark', aslist=False):
	"""Broad spectrum of hex-formatted colors for plotting

	Function to provide a broad palette of colors for plotting, inspired by the
	FlatUI color picker at https://flatuicolors.com/

	Parameters
	----------
	palette : str, optional
		Choice of color palette. Options are `default`, `american`, `aussie`,
		`british`, `canadian`, `chinese`, `french`, `german`, `indian`, `russian`,
		`spanish`, `swedish` and `turkish`.

	shade : str, optional
		`light` or `dark` theme.  Default is `dark`.

	aslist : bool, optional
		Specifies whether to return the colormap as a dictionary,
		e.g. {'k':'#000000', 'w':'#ffffff'} or as a list, e.g. ['#000000', '#ffffff'].
		Default is `False` (returns dictionary)

	Returns
	-------
	colormap : dict or list
		Colormap in list or dictionary format

	Examples
	--------
	>>> cm = ci.colormap()
	>>> print(cm)
	{'b': '#2980b9', 'g': '#27ae60', 'k': '#2c3e50', 'o': '#d35400', 'aq': '#16a085', 
	'p': '#8e44ad', 'r': '#c0392b', 'gy': '#7f8c8d', 'w': '#bdc3c7', 'y': '#f39c12'}
	>>> cm_list = ci.colormap(aslist=True)
	>>> cm_british_light = ci.colormap('british', 'light')

	"""

	pmap = ['default','american','aussie','british','canadian',
			'chinese','dutch','french','german','indian','russian',
			'spanish','swedish','turkish']
	cmap = ['aq','g','b','p','k','y','o','r','w','gy']

	if palette.lower() not in pmap:
		palette = 'default'

	if shade.lower()=='light':
		cl = LIGHT.split('\n')[pmap.index(palette.lower())].split(' ')
	else:
		cl = DARK.split('\n')[pmap.index(palette.lower())].split(' ')

	if aslist:
		return [cl[n] for n in [4,9,7,2,1,6,0,5,3]]

	return {cmap[n]:cl[n] for n in range(len(cl))}

def set_style(style='default', palette='default', shade='dark'):
	"""Preset styles for generating plots

	Adjusts default font sizes, colors and spacing for plots, with presets based 
	on the intended application (i.e. `poster`, `paper`, etc.)

	Parameters
	----------
	style : str, optional
		Preset style based on intended application of the plot. Options are 
		`paper`, `poster`, `presentation`, and the default: `default`.

	palette : str, optional
		Sets colormap palette for plots, see `colormap` for detailed options.

	shade : str, optional
		Sets colormap shade for plots, see `colormap` for detailed options

	Examples
	--------
	>>> ci.set_style('paper')
	>>> ci.set_style('poster', shade='dark')
	>>> ci.set_style('presentation', palette='aussie')

	"""

	from cycler import cycler

	cm = colormap(palette=palette, shade=shade)
	plt.rcParams['axes.prop_cycle'] = cycler(color=[cm['k'],cm['r'],cm['b'],cm['gy'],cm['g'],cm['p'],cm['o'],cm['aq'],cm['y']])

	plt.rcParams['font.family']='sans-serif'
	plt.rcParams['figure.autolayout'] = 'True'
	plt.rcParams['xtick.minor.visible']='True'
	plt.rcParams['ytick.minor.visible']='True'

	plt.rcParams['xtick.major.size']='5'
	plt.rcParams['xtick.minor.size']='3.5'
	plt.rcParams['ytick.major.size']='5'
	plt.rcParams['ytick.minor.size']='3.5'
	
	if style.lower()=='paper':
		plt.rcParams['font.size']='14'
		plt.rcParams['axes.labelsize']='16'
		plt.rcParams['legend.fontsize']='12'

		plt.rcParams['axes.titlepad']='10'
		plt.rcParams['xtick.major.pad']='6'
		plt.rcParams['ytick.major.pad']='6'
		
		plt.rcParams['xtick.major.width']='1.25'
		plt.rcParams['xtick.minor.width']='1.05'
		plt.rcParams['ytick.major.width']='1.25'
		plt.rcParams['ytick.minor.width']='1.05'

		plt.rcParams['axes.linewidth']='1.05'
		
		plt.rcParams['lines.markersize']='4.0'
		plt.rcParams['lines.linewidth'] = '2.2'

		plt.rcParams['errorbar.capsize']='4.5'
		plt.rcParams['lines.markeredgewidth']='1.2'

	elif style.lower()=='poster':
		plt.rcParams['font.size']='16'
		plt.rcParams['axes.labelsize']='18'
		plt.rcParams['legend.fontsize']='14'

		plt.rcParams['axes.titlepad']='12'
		plt.rcParams['xtick.major.pad']='8'
		plt.rcParams['ytick.major.pad']='8'

		plt.rcParams['xtick.major.width']='1.5'
		plt.rcParams['xtick.minor.width']='1.2'
		plt.rcParams['ytick.major.width']='1.5'
		plt.rcParams['ytick.minor.width']='1.2'

		plt.rcParams['axes.linewidth']='1.4'
		
		plt.rcParams['lines.markersize']='4.5'
		plt.rcParams['lines.linewidth'] = '2.8'

		plt.rcParams['errorbar.capsize']='5.0'
		plt.rcParams['lines.markeredgewidth']='1.4'

	elif style.lower()=='presentation':
		plt.rcParams['font.size']='15'
		plt.rcParams['axes.labelsize']='17'
		plt.rcParams['legend.fontsize']='12'

		plt.rcParams['axes.titlepad']='10'
		plt.rcParams['xtick.major.pad']='7'
		plt.rcParams['ytick.major.pad']='7'

		plt.rcParams['xtick.major.width']='1.25'
		plt.rcParams['xtick.minor.width']='1.05'
		plt.rcParams['ytick.major.width']='1.25'
		plt.rcParams['ytick.minor.width']='1.05'

		plt.rcParams['axes.linewidth']='1.15'
		
		plt.rcParams['lines.markersize']='4.0'
		plt.rcParams['lines.linewidth'] = '2.4'

		plt.rcParams['errorbar.capsize']='5.0'
		plt.rcParams['lines.markeredgewidth']='1.4'

	else: ### default
		plt.rcParams['font.size']='14'
		plt.rcParams['axes.labelsize']='15'
		plt.rcParams['legend.fontsize']='11'

		plt.rcParams['axes.titlepad']='8'
		plt.rcParams['xtick.major.pad']='5'
		plt.rcParams['ytick.major.pad']='5'

		plt.rcParams['xtick.major.width']='1.25'
		plt.rcParams['xtick.minor.width']='1.05'
		plt.rcParams['ytick.major.width']='1.25'
		plt.rcParams['ytick.minor.width']='1.05'

		plt.rcParams['axes.linewidth']='1.05'
		
		plt.rcParams['lines.markersize']='3.5'
		plt.rcParams['lines.linewidth'] = '2.0'

		plt.rcParams['errorbar.capsize']='4.2'
		plt.rcParams['lines.markeredgewidth']='1.15'



def _init_plot(**kwargs):
	if any([i in kwargs for i in ['style', 'palette', 'shade']]):
		st = kwargs['style'] if 'style' in kwargs else 'default'
		pa = kwargs['palette'] if 'palette' in kwargs else 'default'
		sh = kwargs['shade'] if 'shade' in kwargs else 'dark'
		set_style(style=st, palette=pa, shade=sh)


	f, ax = None, None
	if 'f' in kwargs and 'ax' in kwargs:
		f, ax = kwargs['f'], kwargs['ax']
	
	N = kwargs['_N_plots'] if '_N_plots' in kwargs else 1

	if f is None or ax is None:
		if 'figsize' in kwargs:
			f, ax = plt.subplots(1, N, figsize=kwargs['figsize'])
		else:
			f, ax = plt.subplots(1, N)

	return f, ax


def _draw_plot(fig, axis, **kwargs):
	f, ax = fig, axis

	if '_default_log' in kwargs:
		if kwargs['_default_log']:
			ax.set_yscale('log')
		else:
			ax.set_yscale('linear')

	if 'scale' in kwargs:
		s = kwargs['scale'].lower()
		if s in ['log','logy']:
			ax.set_yscale('log')
		elif s in ['lin','liny','linear']:
			ax.set_yscale('linear')
		elif s=='logx':
			ax.set_xscale('log')
		elif s=='linx':
			ax.set_xscale('linear')
		elif s in ['linlog','loglog','loglin','linlin']:
			ax.set_xscale(s[:3].replace('lin','linear'))
			ax.set_yscale(s[3:].replace('lin','linear'))

	f.tight_layout()

	if 'saveas' in kwargs:
		if kwargs['saveas'] is not None:
			f.savefig(kwargs['saveas'])


	if 'return_plot' in kwargs:
		if kwargs['return_plot']:
			return f, ax

	if 'show' in kwargs:
		if kwargs['show']:
			plt.show()
	else:
		plt.show()

	

	plt.close()



"""
curie
=====

Provides
  1. Classes for plotting and fitting gamma ray spectra
  2. Access to data from evaluated nuclear reaction libraries
  3. Photon attenuation data
  4. Charged particle stopping power calculations
  5. Generalized Bateman equation solver
  6. Atomic and nuclear structure/decay data

How to use the documentation
----------------------------
Documentation is available in two forms: docstrings provided
with the code, and a users guide located on
`the Curie homepage <https://jtmorrell.github.io/curie/build/html/index.html>`_.

The docstring examples assume that `curie` has been imported as `ci`::

    >>> import curie as ci


Code snippets are indicated by three greater-than signs::

  >>> sp = ci.Spectrum('spectrum.Spe')
  >>> sp.plot()

Use the built-in ``help`` function to view a function or class's docstring::

  >>> help(ci.Spectrum)




"""

from .data import _data_path, download
from .plotting import colormap, set_style
set_style('default')

from .isotope import Isotope
from .compound import Compound, COMPOUND_LIST
from .element import Element

from .spectrum import Spectrum
from .calibration import Calibration

from .decay_chain import DecayChain

from .library import Library
from .reaction import Reaction

from .stack import Stack

__version__ = '0.0.15'
__all__ = ['download', 'colormap', 'set_style', 
          'Isotope', 'Element', 'Compound', 
          'Spectrum', 'Calibration', 'DecayChain', 
          'Library', 'Reaction', 'Stack']